from pathlib import Path

from pydantic import TypeAdapter

from gmol.base.data.mmcif import (
    ChemComp,
    load_mmcif_single,
    mmcif_assemblies,
)


def test_load_heterogeneous_half_occupancy_branch(
    test_data: Path,
    ccd_components: dict[str, ChemComp],
):
    data = load_mmcif_single(test_data / "mmcif" / "4mb4.cif")

    ccd = ccd_components.copy()
    ccd.update(
        TypeAdapter(dict[str, ChemComp]).validate_json(
            (test_data / "ccd" / "components_4mb4.json").read_bytes()
        )
    )

    assemblies = mmcif_assemblies(data, ccd)
    asm = assemblies[0]

    chain_b = asm.chains["B"]

    # NDG discarded by altloc selection logic. Should not overwrite NAG
    assert len(chain_b.branches) == 3
    for branch in chain_b.branches:
        assert branch.ptnr1.comp_id == "NAG"
        assert branch.ptnr2.comp_id == "NAG"

    mmcif_out = asm.to_mmcif("4MB4")
    in_branch_scheme = False
    branch_scheme_lines: list[str] = []

    for line in mmcif_out.splitlines():
        stripped = line.strip()
        if stripped.startswith("_pdbx_branch_scheme."):
            in_branch_scheme = True
            continue
        if (
            in_branch_scheme
            and stripped
            and not stripped.startswith(("_", "#"))
        ):
            branch_scheme_lines.append(stripped)
        elif in_branch_scheme and (stripped.startswith("#") or stripped == ""):
            if branch_scheme_lines:
                break

    assert len(branch_scheme_lines) == 4
    for bsl in branch_scheme_lines:
        tokens = bsl.split()
        mon_id = tokens[3]
        assert mon_id == "NAG"



from pathlib import Path

import pytest
from rdkit import Chem

from gmol.base.wrapper.rdkit import (
    generate_conformer,
    read_mols,
    smi2mol,
    write_mols,
)


@pytest.fixture
def ethane_mol() -> Chem.Mol:
    return smi2mol("CC")


@pytest.fixture
def benzene_mol() -> Chem.Mol:
    return smi2mol("c1ccccc1")


def test_write_mols_single_mol_sdf(tmp_path: Path, ethane_mol: Chem.Mol):
    out = tmp_path / "out.sdf"
    write_mols(out, ethane_mol)
    assert out.is_file()
    mols = list(Chem.SDMolSupplier(str(out)))
    assert len(mols) == 1
    assert mols[0] is not None
    assert Chem.MolToSmiles(mols[0]) == "CC"


def test_write_mols_single_mol_multi_conf_sdf(
    tmp_path: Path,
    ethane_mol: Chem.Mol,
):
    mol = generate_conformer(ethane_mol, ignore_failures=False)

    out = tmp_path / "out.sdf"
    write_mols(out, mol)
    assert out.is_file()
    mols = list(Chem.SDMolSupplier(str(out)))
    assert len(mols) == 1
    assert mols[0] is not None
    assert Chem.MolToSmiles(mols[0]) == "CC"

    mol.AddConformer(mol.GetConformer(), assignId=True)

    write_mols(out, mol)
    assert out.is_file()
    mols = list(Chem.SDMolSupplier(str(out)))
    assert len(mols) == 1
    assert mols[0] is not None
    assert Chem.MolToSmiles(mols[0]) == "CC"


def test_write_mols_single_mol_pdb(tmp_path: Path, ethane_mol: Chem.Mol):
    out = tmp_path / "out.pdb"
    write_mols(out, ethane_mol)
    assert out.is_file()
    content = out.read_text()
    assert "MODEL        1" in content
    assert "ENDMDL" in content


def test_write_mols_single_mol_multi_conf_pdb(
    tmp_path: Path,
    ethane_mol: Chem.Mol,
):
    mol = generate_conformer(ethane_mol, ignore_failures=False)

    out = tmp_path / "out.pdb"
    write_mols(out, mol)
    assert out.is_file()
    content = out.read_text()
    assert content.count("MODEL        1") == 1
    assert content.count("MODEL") == 1
    assert content.count("ENDMDL") == 1

    mol.AddConformer(mol.GetConformer(), assignId=True)

    write_mols(out, mol)
    assert out.is_file()
    content = out.read_text()
    assert content.count("MODEL        1") == 1
    assert content.count("MODEL") == 1
    assert content.count("ENDMDL") == 1


def test_write_mols_list_sdf(
    tmp_path: Path, ethane_mol: Chem.Mol, benzene_mol: Chem.Mol
):
    out = tmp_path / "out.sdf"
    write_mols(out, [ethane_mol, benzene_mol])
    mols = list(Chem.SDMolSupplier(str(out)))
    assert len(mols) == 2
    assert Chem.MolToSmiles(mols[0]) == "CC"
    assert Chem.MolToSmiles(mols[1]) == "c1ccccc1"


def test_write_mols_list_pdb(
    tmp_path: Path, ethane_mol: Chem.Mol, benzene_mol: Chem.Mol
):
    out = tmp_path / "out.pdb"
    write_mols(out, [ethane_mol, benzene_mol])
    content = out.read_text()
    assert "MODEL        1" in content
    assert "MODEL        2" in content
    assert content.count("ENDMDL") == 2
    assert content.count("END") == 3  # 2 ENDMDL + 1 final END


def test_write_mols_skips_none(tmp_path: Path, ethane_mol: Chem.Mol):
    out = tmp_path / "out.sdf"
    write_mols(out, [ethane_mol, None, ethane_mol])
    mols = list(Chem.SDMolSupplier(str(out)))
    assert len(mols) == 2


def test_write_mols_str_path(tmp_path: Path, ethane_mol: Chem.Mol):
    out = tmp_path / "out.sdf"
    write_mols(str(out), ethane_mol)
    assert out.is_file()
    mols = list(Chem.SDMolSupplier(str(out)))
    assert len(mols) == 1


def test_write_mols_unsupported_extension_raises(
    tmp_path: Path, ethane_mol: Chem.Mol
):
    out = tmp_path / "out.xyz"
    with pytest.raises(ValueError, match="Unsupported file extension"):
        write_mols(out, ethane_mol)


def test_write_mols_sdf_kekulize(tmp_path: Path, benzene_mol: Chem.Mol):
    out = tmp_path / "out.sdf"
    write_mols(out, benzene_mol, sdf_kekulize=True)
    assert out.is_file()
    mols = list(Chem.SDMolSupplier(str(out)))
    assert len(mols) == 1
    # Kekulized benzene has explicit single/double bonds
    assert mols[0] is not None
    assert mols[0].GetNumBonds() == 6


LIGAND_NAMES = ["5dwr_ligand", "8a0b_ligand"]


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_read_mols_sanitize(test_data: Path, ligand: str):
    """read_mols with sanitize=True and sanitize=False both return mols."""
    sdf_path = test_data / "ligands" / f"{ligand}.sdf"
    mols_sanitize = read_mols(sdf_path, sanitize=True)
    mols_no_sanitize = read_mols(sdf_path, sanitize=False)
    assert len(mols_sanitize) == len(mols_no_sanitize)
    assert len(mols_sanitize) >= 1
    for a, b in zip(mols_sanitize, mols_no_sanitize, strict=True):
        assert a.GetNumHeavyAtoms() == b.GetNumHeavyAtoms()


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_read_mols_remove_h(test_data: Path, ligand: str):
    """read_mols with remove_h=True returns mols with no more atoms than remove_h=False."""
    sdf_path = test_data / "ligands" / f"{ligand}.sdf"
    mols_with_h = read_mols(sdf_path, remove_h=False)
    mols_no_h = read_mols(sdf_path, remove_h=True)
    assert len(mols_with_h) == len(mols_no_h)
    assert len(mols_with_h) >= 1
    for with_h, no_h in zip(mols_with_h, mols_no_h, strict=True):
        assert no_h.GetNumAtoms() <= with_h.GetNumAtoms()


def test_read_mols_multiple_mol2(test_data: Path):
    mol2_path = test_data / "ligands" / "ligands.mol2"
    mols = read_mols(mol2_path)
    assert len(mols) == 2

    assert mols[0].GetNumAtoms() == 56
    assert mols[0].GetNumBonds() == 59

    assert mols[1].GetNumAtoms() == 22
    assert mols[1].GetNumBonds() == 25


def test_read_mols_multiple_pdb(test_data: Path):
    pdb_path = test_data / "ligands" / "ligands.pdb"
    mols = read_mols(pdb_path)
    assert len(mols) == 2

    assert mols[0].GetNumAtoms() == 56
    assert mols[0].GetNumBonds() == 59

    assert mols[1].GetNumAtoms() == 22
    assert mols[1].GetNumBonds() == 25


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_write_mols_ligands_sdf(tmp_path: Path, test_data: Path, ligand: str):
    """Read SDF from ligands dir, write to tmp SDF, read back and compare count."""
    sdf_path = test_data / "ligands" / f"{ligand}.sdf"
    mols = read_mols(sdf_path)
    assert len(mols) >= 1

    out = tmp_path / "out.sdf"
    write_mols(out, mols)
    assert out.is_file()
    roundtrip = list(Chem.SDMolSupplier(str(out)))
    roundtrip = [m for m in roundtrip if m is not None]
    assert len(roundtrip) == len(mols)
    for a, b in zip(mols, roundtrip, strict=True):
        assert a.GetNumHeavyAtoms() == b.GetNumHeavyAtoms()


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_write_mols_ligands_multi_sdf(
    tmp_path: Path, test_data: Path, ligand: str
):
    """Read multi-molecule SDF from ligands dir, write and roundtrip."""
    sdf_path = test_data / "ligands" / f"{ligand}.sdf"
    mols = read_mols(sdf_path)
    assert len(mols) >= 1

    out = tmp_path / "out.sdf"
    multi_confs = mols * 2
    write_mols(out, multi_confs)
    assert out.is_file()
    roundtrip = list(Chem.SDMolSupplier(str(out)))
    roundtrip = [m for m in roundtrip if m is not None]
    assert len(roundtrip) == len(multi_confs)


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_write_mols_ligands_pdb_roundtrip(
    tmp_path: Path, test_data: Path, ligand: str
):
    """Read PDB from ligands dir, write to tmp PDB, check structure and model count."""
    pdb_path = test_data / "ligands" / f"{ligand}.pdb"
    mols = read_mols(pdb_path)
    assert len(mols) >= 1

    out = tmp_path / "out.pdb"
    write_mols(out, mols)
    assert out.is_file()
    content = out.read_text()
    assert "MODEL        1" in content
    assert "ENDMDL" in content
    assert content.count("MODEL") == len(mols)
    assert content.count("ENDMDL") == len(mols)


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_write_mols_ligands_from_mol2_to_sdf(
    tmp_path: Path, test_data: Path, ligand: str
):
    """Read MOL2 from ligands dir, write as SDF, verify mol count."""
    mol2_path = test_data / "ligands" / f"{ligand}.mol2"
    mols = read_mols(mol2_path)
    assert len(mols) >= 1

    out = tmp_path / "out.sdf"
    write_mols(out, mols)
    assert out.is_file()
    written = list(Chem.SDMolSupplier(str(out)))
    written = [m for m in written if m is not None]
    assert len(written) == len(mols)


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_write_mols_ligands_from_mol2_to_pdb(
    tmp_path: Path, test_data: Path, ligand: str
):
    """Read MOL2 from ligands dir, write as PDB, verify model count."""
    mol2_path = test_data / "ligands" / f"{ligand}.mol2"
    mols = read_mols(mol2_path)
    assert len(mols) >= 1

    out = tmp_path / "out.pdb"
    write_mols(out, mols)
    assert out.is_file()
    content = out.read_text()
    assert content.count("MODEL") == len(mols)
    assert content.count("ENDMDL") == len(mols)


@pytest.mark.parametrize("ligand", LIGAND_NAMES)
def test_write_mols_ligands_skips_none(
    tmp_path: Path, test_data: Path, ligand: str
):
    """write_mols skips None when list is built from ligand data."""
    sdf_path = test_data / "ligands" / f"{ligand}.sdf"
    mols = read_mols(sdf_path)
    assert len(mols) >= 1
    mixed = [mols[0], None, mols[0]]

    out = tmp_path / "out.sdf"
    write_mols(out, mixed)
    written = list(Chem.SDMolSupplier(str(out)))
    written = [m for m in written if m is not None]
    assert len(written) == 2



# ruff: noqa: A001,INP001

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import re

from pydantic import BaseModel
from sphinx.ext import autodoc

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "gmol-base"
copyright = "2026 SNU Compbio Lab"
author = "SNU Compbio Lab"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autodoc_member_order = "bysource"
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
    "ignore-module-all": True,
}
autodoc_inherit_docstrings = False

autosummary_generate = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "pydantic": ("https://docs.pydantic.dev/latest", None),
    "rdkit": ("https://www.rdkit.org/docs", None),
    "nurikit": ("https://nurikit.readthedocs.io/latest", None),
}

add_module_names = False
python_display_short_literal_types = True
python_use_unqualified_type_names = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/seoklab/gmol-base",
            "icon": "fa-brands fa-github",
        }
    ],
    "secondary_sidebar_items": ["page-toc"],
    "footer_start": ["last-updated"],
    "footer_center": ["copyright"],
    "footer_end": ["sphinx-version", "theme-version"],
}

html_title = project
html_static_path = ["_static"]
html_last_updated_fmt = "%Y-%m-%d %H:%M"

mathjax_path = "https://cdn.jsdelivr.net/npm/mathjax@4/tex-mml-chtml.js"


class StrippingDocumenter(autodoc.ClassDocumenter):
    _strip_re = re.compile(r"\s*Bases:\s*:py:class:`(object)`\s*")

    def add_line(self, line: str, source: str, *lineno: int):
        if self._strip_re.fullmatch(line):
            return
        super().add_line(line, source, *lineno)

    def get_object_members(self, want_all: bool):
        flag, members = super().get_object_members(want_all)

        if isinstance(self.object, type) and issubclass(
            self.object, BaseModel
        ):
            members = [
                member
                for member in members
                if not hasattr(BaseModel, member.__name__)
            ]

        return flag, members


autodoc.ClassDocumenter = StrippingDocumenter  # type: ignore[misc]



import datetime as dt
from pathlib import Path

import numpy as np
import pytest
from pydantic import TypeAdapter

from gmol.base.data.mmcif import (
    ChemComp,
    MolType,
    filter_mmcif,
    load_mmcif_single,
    mmcif_assemblies,
)
from gmol.base.data.mmcif.input import (
    ExtraBondPartner,
    build_input,
    chain_type_to_constants,
)


@pytest.fixture(scope="session")
def ccd(test_data: Path):
    return TypeAdapter(dict[str, ChemComp]).validate_json(
        test_data.joinpath("ccd", "components_min.json").read_bytes()
    )


def _load_input(
    pdb_id: str,
    test_data: Path,
    ccd: dict[str, ChemComp],
    split: bool,
):
    mmcif = test_data / "mmcif" / f"{pdb_id}.cif"
    data = load_mmcif_single(mmcif)

    assemblies = mmcif_assemblies(data, ccd)
    assert len(assemblies) == 1

    result = filter_mmcif(
        assemblies[0], ccd, cutoff_date=dt.date(9999, 12, 31)
    )
    assert result is not None

    input_data = build_input(result, ccd, split_modified=split)
    return input_data


def test_build_input_basic_transforms(
    test_data: Path,
    ccd: dict[str, ChemComp],
):
    input_data = _load_input("4hf7", test_data, ccd, False)

    # test ARG swap (dist (CD, NH1) < dist(CD, NH2))
    sample_protein = input_data.polymers[0]
    sample_arg_idx = 10
    arg_coords = sample_protein.atom_coords[sample_arg_idx]

    arg_atom_idxs = chain_type_to_constants[MolType.Protein].residue_atom_idxs[
        "ARG"
    ]
    CD_idx, NH1_idx, NH2_idx = (
        arg_atom_idxs["CD"],
        arg_atom_idxs["NH1"],
        arg_atom_idxs["NH2"],
    )

    dist_CD_NH1 = np.linalg.norm(arg_coords[CD_idx] - arg_coords[NH1_idx])
    dist_CD_NH2 = np.linalg.norm(arg_coords[CD_idx] - arg_coords[NH2_idx])

    assert dist_CD_NH1 < dist_CD_NH2, (
        f"ARG residue at index {sample_arg_idx}: "
        f"NH1 ({dist_CD_NH1:.3f}) should be closer to CD than NH2 ({dist_CD_NH2:.3f})."
    )

    # test MSE restype is changed to MET
    sample_mse_idx = 31

    mse_chem_comp = sample_protein.chem_comp_raw[sample_mse_idx]
    assert mse_chem_comp == "MSE", (
        f"MSE residue index {sample_mse_idx} should have chem_comp_id 'MSE'"
    )

    mse_restype = sample_protein.restype[sample_mse_idx]
    met_restype = chain_type_to_constants[
        MolType.Protein
    ].restype_order_with_x["M"]
    assert mse_restype == met_restype, (
        f"MSE residue index {sample_mse_idx} should have restype {met_restype}"
    )

    # test for modified residue: it treat as a covalent-lig (bonded with protein chain)
    modres_chain_id = "A_1-MOD34"
    prev_residue = ("A_1", 33, "C")
    next_residue = ("A_1", 35, "N")

    bond_pairs = {
        frozenset({prev_residue, (modres_chain_id, 0, "N")}),
        frozenset({next_residue, (modres_chain_id, 0, "C")}),
    }
    total_extra_bonds = {
        frozenset(
            {
                (bond.src.chain_id, bond.src.res_idx, bond.src.atom_id),
                (bond.dst.chain_id, bond.dst.res_idx, bond.dst.atom_id),
            }
        )
        for bond in input_data.extra_bonds
    }

    assert bond_pairs.issubset(total_extra_bonds), (
        f"Modified residue ({modres_chain_id}) is connected to previous residue {prev_residue} and next residue {next_residue} correctly."
    )


def test_build_input_equivalent_entity_different_chain(
    test_data: Path,
    ccd: dict[str, ChemComp],
):
    input_data = _load_input("7rtb", test_data, ccd, False)

    modres = [
        lig for lig in input_data.ligands if lig.chain_id.split("-")[0] == "F"
    ]

    modres_entities = {lig.entity_id for lig in modres}
    assert len(modres_entities) == 1, (
        "Same modres should have single entity id"
    )

    modres_chains = {lig.chain_id for lig in modres}
    assert len(modres_chains) == 2, "Same modres should have two chain ids"

    input_data = _load_input("7rtb", test_data, ccd, True)

    modres = [
        lig for lig in input_data.ligands if lig.chain_id.split("-")[0] == "F"
    ]

    modres_entities = {lig.entity_id for lig in modres}
    assert len(modres_entities) == 2, (
        "Must have two splitted side chain entities"
    )

    modres_chains = {lig.chain_id for lig in modres}
    assert len(modres_chains) == 4, "Must have 2 x 2 splitted side chains"


def test_build_input_disulfide_covalent_ligand(
    test_data: Path,
    ccd: dict[str, ChemComp],
):
    input_data = _load_input("7qgw", test_data, ccd, False)

    bonds = {
        tuple(sorted([bond.src, bond.dst])): bond
        for bond in input_data.extra_bonds
    }

    assert (
        ExtraBondPartner(chain_id="A", res_idx=21, atom_id="SG"),
        ExtraBondPartner(chain_id="A", res_idx=64, atom_id="SG"),
    ) in bonds, "Missing disulfide bond between CYS 22.A and CYS 64.A"

    assert (
        ExtraBondPartner(chain_id="A", res_idx=24, atom_id="SG"),
        ExtraBondPartner(chain_id="K", res_idx=0, atom_id="C25"),
    ) in bonds, "Missing covalent bond between CYS 25.A and RN2 309.A"



"""This module is written to follow the syntax rules in the reference below.

Reference:
    CIF 1.1 syntax (general):
    https://www.iucr.org/resources/cif/spec/version1.1/cifsyntax#general
"""

from collections.abc import Sequence


def mmcif_bond_order(order: int) -> str:
    return {1: "sing", 2: "doub", 3: "trip", 4: "quad"}[order]


def mmcif_bool(pred: bool) -> str:
    return "Y" if pred else "N"


_RESERVED_PREFIXES = ("data_", "save_")
_RESERVED_EXACT = frozenset({"loop_", "stop_", "global_"})
_RESERVED_START_CHARS = frozenset("_#$[];'\"")


def _mmcif_text_field(v: str) -> str:
    normalized = v.replace("\r\n", "\n").replace("\r", "\n")
    if "\n;" in normalized:
        raise ValueError(
            "Cannot encode a CIF text field containing a line that starts with ';'"
        )
    return f";{normalized}\n;"


def _needs_quoted(v: str) -> bool:
    if not v:
        return True

    lower = v.lower()

    return (
        any(c.isspace() for c in v)
        or v[0] in _RESERVED_START_CHARS
        or lower in _RESERVED_EXACT
        or lower.startswith(_RESERVED_PREFIXES)
    )


def _mmcif_quote(v: str) -> str:
    if "\n" in v or "\r" in v:
        return _mmcif_text_field(v)

    if "'" not in v:
        return f"'{v}'"
    if '"' not in v:
        return f'"{v}"'
    return _mmcif_text_field(v)


def _mmcif_escape(v: str) -> str:
    if _needs_quoted(v):
        return _mmcif_quote(v)
    return v


def _assert_join(row: Sequence[str | int], num_fields: int) -> str:
    if len(row) != num_fields:
        raise ValueError("Row length does not match field count")

    tokens = [_mmcif_escape(str(c)) for c in row]
    if any("\n" in tok for tok in tokens):
        return "\n".join(tokens)
    return " ".join(tokens)


def mmcif_write_block(
    block: str,
    fields: Sequence[str],
    data: Sequence[Sequence[str | int]],
) -> str:
    if not fields:
        raise ValueError("Fields must not be empty")

    if not data:
        return ""

    content = """
#
loop_
""" + "\n".join(f"_{block}.{field}" for field in fields)

    content += "\n" + "\n".join(_assert_join(row, len(fields)) for row in data)

    return content



import logging
import re
from collections.abc import Iterable
from itertools import count
from pathlib import Path
from typing import cast

from rdkit import Chem
from rdkit.Chem import AllChem

_logger = logging.getLogger(__name__)


def smi2mol(
    smiles: str,
    kekuilze: bool = False,
    sanitize: bool = True,
) -> Chem.Mol:
    m = Chem.MolFromSmiles(smiles, sanitize=sanitize)
    if m is None:  # pyright: ignore[reportUnnecessaryComparison]
        raise ValueError(f"Fail to convert smiles ({smiles}) to RDKit Mol.")
    if kekuilze:
        Chem.Kekulize(m)
    return m


def mol2smi(mol: Chem.Mol, kekuleSmiles: bool = True) -> str:
    return Chem.MolToSmiles(mol, kekuleSmiles)


def get_all_bonds(mol: Chem.Mol) -> list[tuple[int, int]]:
    return [(t.GetBeginAtomIdx(), t.GetEndAtomIdx()) for t in mol.GetBonds()]


def find_all_bonds_with_atom_idx(
    mol: Chem.Mol, atom_indices: list[int]
) -> list[tuple[int, int]]:
    """
    Find all chemical bonds in the molecule given atom indices.
    """
    all_bonds = get_all_bonds(mol)
    result = set()
    for bond in all_bonds:
        b_src, b_dst = bond
        if b_src in atom_indices and b_dst in atom_indices:
            result.add(tuple(sorted(bond)))
    return list(result)  # type: ignore[arg-type]


def generate_conformer(
    mol: Chem.Mol,
    max_attempts: int = 5,
    optimize_iters: int = 500,
    ignore_failures: bool = True,
) -> Chem.Mol:
    # To preserve original conformer
    orig_m = Chem.Mol(mol)

    try:
        new_m = Chem.Mol(mol)
        new_m.RemoveAllConformers()
        new_m = Chem.AddHs(new_m)

        # Attempt embedding
        params = AllChem.ETKDGv3()
        params.maxIterations = max_attempts
        success = False

        if AllChem.EmbedMolecule(new_m, params) == 0:
            success = True

        if not success:
            _logger.warning(
                "Initial embedding failed, retrying with random coordinates."
            )
            params.useRandomCoords = True
            if AllChem.EmbedMolecule(new_m, params) != 0:
                raise ValueError(
                    "Conformer embedding failed after randomization."
                )

        # MMFF optimization
        opt_status = AllChem.MMFFOptimizeMolecule(
            new_m,
            confId=0,
            mmffVariant="MMFF94s",
            maxIters=optimize_iters,
        )
        if opt_status != 0:
            _logger.warning(
                "MMFF optimization returned status %s.", opt_status
            )

        # Assign new stereochemistry for featurizers
        result = Chem.RemoveHs(new_m)
        Chem.rdmolops.AssignStereochemistryFrom3D(result)
        return result

    except Exception:
        if ignore_failures:
            _logger.exception(
                "Generating conformer failed. Use the original conformer."
            )
            return orig_m

        raise


def _split_read_mol2(path: str):

    def _read_mol(block: list[str]):
        mol2 = "".join(block)
        mol = Chem.MolFromMol2Block(
            mol2,
            sanitize=False,
            removeHs=False,
        )
        return cast(Chem.Mol | None, mol)

    with open(path) as f:
        block: list[str] = []

        for line in f:
            if line.startswith("@<TRIPOS>MOLECULE") and block:
                mol = _read_mol(block)
                if mol is not None:
                    yield mol

                block.clear()

            block.append(line)

        if block:
            mol = _read_mol(block)
            if mol is not None:
                yield mol


def _split_read_pdb(path: str):

    def _read_mol(block: list[str]):
        pdb = "".join(block)
        mol = Chem.MolFromPDBBlock(
            pdb,
            sanitize=False,
            removeHs=False,
        )
        return cast(Chem.Mol | None, mol)

    with open(path) as f:
        block: list[str] = []

        for line in f:
            if line.startswith("MODEL"):
                block.clear()
                continue

            if line.startswith("ENDMDL"):
                mol = _read_mol(block)
                if mol is not None:
                    yield mol

                block.clear()
                continue

            block.append(line)

        if block:
            # If no "MODEL/ENDMDL" is found, the whole file is interpreted as a
            # single molecule.
            mol = _read_mol(block)
            if mol is not None:
                yield mol


def read_mols(
    file_path: Path | str,
    sanitize: bool = True,
    remove_h: bool = False,
) -> list[Chem.Mol]:
    """Read a molecular file into a list of RDKit Mol objects.

    Supported formats: ``.mol2``, ``.sdf``, ``.pdb``.

    :param file_path: Path to the molecular file.
    :param sanitize: If True, sanitize each molecule after reading.
    :param remove_h: If True, remove explicit hydrogens from each molecule.
    :returns: List of RDKit Mol objects; entries that failed to parse are omitted.
    """
    file_path = Path(file_path)

    if not file_path.exists():
        raise FileNotFoundError(f"File not found: {file_path}")

    ext = file_path.suffix.lower()
    if ext not in (".mol2", ".sdf", ".pdb"):
        raise ValueError(
            f"Unsupported file extension '{ext}'. Supported formats: .mol2, .sdf, .pdb."
        )

    if ext == ".mol2":
        mols = list(_split_read_mol2(str(file_path)))
    elif ext == ".sdf":
        with Chem.SDMolSupplier(
            str(file_path), sanitize=False, removeHs=False
        ) as suppl:
            mols = [m for m in suppl if cast(Chem.Mol | None, m) is not None]
    elif ext == ".pdb":
        mols = list(_split_read_pdb(str(file_path)))

    if sanitize:
        for mol in mols:
            Chem.SanitizeMol(mol)

    if remove_h:
        mols = [Chem.RemoveHs(mol, sanitize=sanitize) for mol in mols]

    return mols


_end_re = re.compile(r"^END\s+", re.MULTILINE)


def write_mols(
    save_path: Path | str,
    mols: Chem.Mol | None | Iterable[Chem.Mol | None],
    sdf_kekulize: bool = False,
) -> None:
    """Write a list of RDKit Mol objects to a file.

    Supported extensions are ``.sdf`` and ``.pdb``.

    :param save_path: Path to the output file.
    :param mols: Molecule or list of molecules to write; None entries are skipped.
    :param sdf_kekulize: If True, kekulize molecules when writing to an SDF file.
    """
    if mols is None:
        return

    if isinstance(mols, Chem.Mol):
        mols = [mols]

    save_path = Path(save_path)
    ext = save_path.suffix.lower()

    if ext == ".sdf":
        with Chem.SDWriter(str(save_path)) as w:
            w.SetKekulize(sdf_kekulize)
            for m in mols:
                if m is not None:
                    w.write(m)
    elif ext == ".pdb":
        with save_path.open("w") as f:
            written = False
            model_num = count(1)
            for m in mols:
                if m is not None:
                    written = True
                    idx = next(model_num)

                    # rdkit writes all conformers to pdb by default, but we
                    # only want to write single conformer per molecule.
                    # If there are no conformers, confId=-1 will be used to
                    # write the molecule without 3D coordinates.
                    confid = -1
                    if m.GetNumConformers() > 1:
                        confid = m.GetConformer().GetId()

                    pdb_block = Chem.MolToPDBBlock(m, confId=confid)
                    pdb_block = _end_re.sub("", pdb_block)
                    f.write(f"MODEL {idx:8d}\n")
                    f.write(pdb_block)
                    f.write("ENDMDL\n")

            if written:
                f.write("END\n")
    else:
        raise ValueError(
            f"Unsupported file extension '{ext}'. Supported: .sdf, .pdb"
        )



"""
Some of the code is imported from DeepMind's AlphaFold.
See https://github.com/google-deepmind/alphafold/blob/main/alphafold/data/tools/hhsearch.py
for the original code.

.. admonition:: License
    :collapsible: closed

    .. code-block::

        Copyright 2021 DeepMind Technologies Limited

        Licensed under the Apache License, Version 2.0 (the "License");
        you may not use this file except in compliance with the License.
        You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

        Unless required by applicable law or agreed to in writing, software
        distributed under the License is distributed on an "AS IS" BASIS,
        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
        See the License for the specific language governing permissions and
        limitations under the License.
"""

import logging
import subprocess as sp
from pathlib import Path
from typing import ClassVar

_logger = logging.getLogger(__name__)


class HHSearch:
    """Python wrapper of the HHsearch binary."""

    output_format: ClassVar[str] = "hhr"
    input_format: ClassVar[str] = "a3m"

    def __init__(
        self,
        *,
        binary_path: Path,
        databases: list[Path],
        output_dir: Path,
        maxseq: int = 1_000_000,
    ):
        """Initializes the Python HHsearch wrapper.

        Args:
          binary_path: The path to the HHsearch executable.
          databases: A sequence of HHsearch database paths. This should be the
            common prefix for the database files (i.e. up to but not including
            _hhm.ffindex etc.)
          maxseq: The maximum number of rows in an input alignment. Note that this
            parameter is only supported in HHBlits version 3.1 and higher.
        """
        self.binary_path = binary_path
        self.databases = databases
        self.maxseq = maxseq
        self.output_dir = output_dir

        for database_path in self.databases:
            if not list(database_path.parent.glob(f"{database_path.name}_*")):
                raise ValueError(
                    f"Could not find HHsearch database {database_path}"
                )

        self.output_dir.mkdir(exist_ok=True, parents=True)

    def query(self, a3m: Path):
        """Queries the database using HHsearch using a given a3m."""
        db_cmd: list[str | Path] = []
        for db in self.databases:
            db_cmd.append("-d")
            db_cmd.append(db)

        # fmt: off
        cmd: list[str | Path] = [
            self.binary_path,
            "-i", a3m,
            "-o", self.output_dir / "output.hhr",
            "-maxseq", str(self.maxseq),
            *db_cmd,
        ]
        # fmt: on

        _logger.info('Launching subprocess "%s"', cmd)
        sp.run(cmd, check=True)



"""
Some of the code is imported from OpenFold.
See https://github.com/aqlaboratory/openfold/blob/main/openfold/np/residue_constants.py
for the original code.

.. admonition:: License
    :collapsible: closed

    .. code-block::

        Copyright 2021 AlQuraishi Laboratory
        Copyright 2021 DeepMind Technologies Limited

        Licensed under the Apache License, Version 2.0 (the "License");
        you may not use this file except in compliance with the License.
        You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

        Unless required by applicable law or agreed to in writing, software
        distributed under the License is distributed on an "AS IS" BASIS,
        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
        See the License for the specific language governing permissions and
        limitations under the License.
"""

# ===same ordering of Amino acid as in AlphaFold2======
# This is the standard residue order when coding AA type as a number.
# Reproduce it by taking 3-letter AA codes and sorting them alphabetically.
aa_restypes = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "Q",
    "E",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]

aa_restype_order = {restype: i for i, restype in enumerate(aa_restypes)}
aa_restype_order_with_x = {
    **aa_restype_order,
    "X": len(aa_restypes),
}
aa_restype_num = len(aa_restypes)
aa_unk_restype_index = aa_restype_num

aa_restypes_with_x = [*aa_restypes, "X"]
aa_restype_order_with_x = {
    restype: i for i, restype in enumerate(aa_restypes_with_x)
}

aa_restype_1to3 = {
    "A": "ALA",
    "R": "ARG",
    "N": "ASN",
    "D": "ASP",
    "C": "CYS",
    "Q": "GLN",
    "E": "GLU",
    "G": "GLY",
    "H": "HIS",
    "I": "ILE",
    "L": "LEU",
    "K": "LYS",
    "M": "MET",
    "F": "PHE",
    "P": "PRO",
    "S": "SER",
    "T": "THR",
    "W": "TRP",
    "Y": "TYR",
    "V": "VAL",
}

aa_restype_3to1 = {v: k for k, v in aa_restype_1to3.items()}

aa_restype_to_atom14 = {
    "ALA": ["N", "CA", "C", "O", "CB", "", "", "", "", "", "", "", "", ""],
    "ARG": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD",
        "NE",
        "CZ",
        "NH1",
        "NH2",
        "",
        "",
        "",
    ],
    "ASN": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "OD1",
        "ND2",
        "",
        "",
        "",
        "",
        "",
        "",
    ],
    "ASP": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "OD1",
        "OD2",
        "",
        "",
        "",
        "",
        "",
        "",
    ],
    "CYS": ["N", "CA", "C", "O", "CB", "SG", "", "", "", "", "", "", "", ""],
    "GLN": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD",
        "OE1",
        "NE2",
        "",
        "",
        "",
        "",
        "",
    ],
    "GLU": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD",
        "OE1",
        "OE2",
        "",
        "",
        "",
        "",
        "",
    ],
    "GLY": ["N", "CA", "C", "O", "", "", "", "", "", "", "", "", "", ""],
    "HIS": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "ND1",
        "CD2",
        "CE1",
        "NE2",
        "",
        "",
        "",
        "",
    ],
    "ILE": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG1",
        "CG2",
        "CD1",
        "",
        "",
        "",
        "",
        "",
        "",
    ],
    "LEU": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD1",
        "CD2",
        "",
        "",
        "",
        "",
        "",
        "",
    ],
    "LYS": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD",
        "CE",
        "NZ",
        "",
        "",
        "",
        "",
        "",
    ],
    "MET": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "SD",
        "CE",
        "",
        "",
        "",
        "",
        "",
        "",
    ],
    "PHE": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD1",
        "CD2",
        "CE1",
        "CE2",
        "CZ",
        "",
        "",
        "",
    ],
    "PRO": ["N", "CA", "C", "O", "CB", "CG", "CD", "", "", "", "", "", "", ""],
    "SER": ["N", "CA", "C", "O", "CB", "OG", "", "", "", "", "", "", "", ""],
    "THR": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "OG1",
        "CG2",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
    ],
    "TRP": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD1",
        "CD2",
        "NE1",
        "CE2",
        "CE3",
        "CZ2",
        "CZ3",
        "CH2",
    ],
    "TYR": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD1",
        "CD2",
        "CE1",
        "CE2",
        "CZ",
        "OH",
        "",
        "",
    ],
    "VAL": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG1",
        "CG2",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
    ],
}



import datetime as dt

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.spatial import KDTree

from gmol.base.const import (
    CRYSTALLIZATION_AIDS,
    CRYSTALLOGRAPHY_METHODS,
    IONS,
    LIGAND_EXCLUSION_LIST,
    aa_restype_3to1,
)
from .assembly import Assembly, AssemblyAtom, MolType, Residue
from .parse import ChemComp

__all__ = ["filter_mmcif", "filter_structure"]


def new_atom_mask(assembly: Assembly) -> NDArray[np.bool_]:
    return np.zeros(len(assembly.atoms), dtype=bool)


def remove_min_resolved_chains(
    assembly: Assembly,
    min_resolved: int = 4,
):
    mask = new_atom_mask(assembly)

    for chain in assembly.chains_of_type(MolType.Protein):
        if len(chain.residue_ids) < min_resolved:
            mask[chain.atom_idxs] = True
    return mask


def remove_waters(assembly: Assembly) -> NDArray[np.bool_]:
    return np.isin(assembly.comp_ids, ["HOH", "DOD", "DOH"])


def remove_all_unknown_chains(assembly: Assembly):
    mask = new_atom_mask(assembly)
    for chain in assembly.chains_of_type(MolType.Protein):
        if all(
            residue.chem_comp.id not in aa_restype_3to1
            for residue in assembly.residues_of_chain(chain)
        ):
            mask[chain.atom_idxs] = True
    return mask


def _find_clashes(
    coords: NDArray[np.float64], cutoff: float = 1.7
) -> NDArray[np.int64]:
    kdtree = KDTree(coords)
    neighbors = kdtree.query_pairs(cutoff, output_type="ndarray")
    return neighbors


def _exclude_metals(
    atom_comps: NDArray[np.str_],
    clashes: NDArray[np.int64],
):
    clashing_comps = atom_comps[clashes]
    has_metal = np.isin(clashing_comps, list(IONS))
    return clashes[~has_metal.any(axis=1)]


def _per_chain_clashes_ratio(
    atoms: list[AssemblyAtom],
    chain_lengths: pd.Series,
    clashes: NDArray[np.int64],
) -> pd.DataFrame:
    atom_chains = np.array([atom.chain_id for atom in atoms], dtype=np.str_)
    clashing_chains = atom_chains[clashes]
    clashing_chains = clashing_chains[
        clashing_chains[:, 0] != clashing_chains[:, 1]
    ]

    df = pd.DataFrame(clashing_chains)
    df = df.groupby(by=0, sort=False, as_index=False).value_counts(
        sort=False, dropna=False
    )
    df = pd.concat([df, df.rename(columns={0: 1, 1: 0})], axis=0).set_index(0)
    df["ratio"] = df["count"] / chain_lengths.loc[df.index].array
    return df


def _chain_to_remove(
    df: pd.DataFrame,
    chain_lengths: pd.Series,
    cutoff: float,
) -> str | None:
    high_clash_cnts = (df["ratio"] > cutoff).groupby(level=0, sort=False).sum()

    max_clash_cnt = high_clash_cnts.max()
    if not max_clash_cnt:
        return None

    candidates = high_clash_cnts[high_clash_cnts == max_clash_cnt].index
    if len(candidates) == 1:
        return candidates[0]

    candidate_stats = df[df.index.isin(candidates)]
    ratio_sum = candidate_stats["ratio"].groupby(level=0, sort=False).sum()

    candidates = ratio_sum[ratio_sum >= ratio_sum.max() - 1e-3].index
    if len(candidates) == 1:
        return candidates[0]

    lengths = chain_lengths.loc[candidates].sort_index(ascending=False)
    return lengths.idxmin()


def remove_clashing_chains(
    assembly: Assembly,
    clash_threshold: float = 1.7,
    clash_cutoff: float = 0.3,
):
    mask = new_atom_mask(assembly)

    clashes = _find_clashes(assembly.coords, clash_threshold)
    clashes = _exclude_metals(assembly.comp_ids, clashes)

    ch_lengths = pd.Series(
        [len(ch.atom_idxs) for ch in assembly.chains.values()],
        index=list(assembly.chains.keys()),
    )

    df = _per_chain_clashes_ratio(assembly.atoms, ch_lengths, clashes)
    while True:
        if df.empty:
            break

        ch = _chain_to_remove(df, ch_lengths, clash_cutoff)
        if ch is None:
            break

        mask[assembly.chains[ch].atom_idxs] = True
        df = df[(df.index != ch) & (df[1] != ch)]

    return mask


def remove_non_ccd_atoms(
    assembly: Assembly,
    chem_comp_dict: dict[str, ChemComp],
):
    mask = new_atom_mask(assembly)

    removed_atoms: list[int] = []
    for chain in assembly.chains.values():
        for residue in assembly.residues_of_chain(chain):
            result = chem_comp_dict.get(residue.chem_comp.id, None)
            if result is None:
                continue

            ccd_atoms: set[str] = set([atom.atom_id for atom in result.atoms])
            removed_atoms.extend(
                atom.atom_idx
                for atom in assembly.atoms_of_residue(residue)
                if atom.atom_id not in ccd_atoms
            )

    mask[removed_atoms] = True
    return mask


def remove_large_ca_distance(
    assembly: Assembly,
    max_distance: float = 10,
):
    mask = new_atom_mask(assembly)

    for chain in assembly.chains_of_type(MolType.Protein):
        chain_ca_coords = np.full(
            (len(chain.seqres), 3), np.nan, dtype=np.float64
        )

        indices: list[int] = []
        known_ca_coords: list[NDArray[np.float64]] = []
        for i, seqres in enumerate(chain.seqres):
            if seqres.res_id is None:
                continue

            res = assembly.residues[seqres.res_id]
            for atom in assembly.atoms_of_residue(res):
                if atom.atom_id == "CA":
                    indices.append(i)
                    known_ca_coords.append(assembly.coords[atom.atom_idx])
                    break

        if not known_ca_coords:
            continue

        chain_ca_coords[indices] = known_ca_coords

        pairwise_dists = np.linalg.norm(
            chain_ca_coords[1:] - chain_ca_coords[:-1],
            axis=1,
        )
        consec_dists = pairwise_dists[~np.isnan(pairwise_dists)]
        if np.any(consec_dists > max_distance):
            mask[chain.atom_idxs] = True

    return mask


def ligand_atom_mask(assembly: Assembly) -> NDArray[np.bool_]:
    mask = new_atom_mask(assembly)
    for chain in assembly.chains_of_type(MolType.Ligand):
        mask[chain.atom_idxs] = True
    return mask


def remove_crystallization_aids(assembly: Assembly, method: str):
    if method not in CRYSTALLOGRAPHY_METHODS:
        return new_atom_mask(assembly)

    return np.isin(assembly.comp_ids, list(CRYSTALLIZATION_AIDS))


def remove_excluded_ligands(assembly: Assembly):
    return np.isin(assembly.comp_ids, list(LIGAND_EXCLUSION_LIST))


def remove_ions(assembly: Assembly):
    return np.isin(assembly.comp_ids, list(IONS))


def _is_ligand(chem_comp: ChemComp) -> bool:
    return all(
        chem_type not in chem_comp.type
        for chem_type in ["peptide", "rna", "dna"]
    )


def _find_leaving_atoms(
    assembly: Assembly,
    residue: Residue,
    ccd_result: ChemComp,
):
    leaving_atoms_table = {
        cc_atom.atom_id: cc_atom.pdbx_leaving_atom_flag
        for cc_atom in ccd_result.atoms
    }

    leaving_atoms: list[int] = []
    for aid, leaving_flag in leaving_atoms_table.items():
        if not leaving_flag:
            continue

        for atom in assembly.atoms_of_residue(residue):
            if atom.atom_id == aid:
                leaving_atoms.append(atom.atom_idx)

    if len(leaving_atoms) != len(set(leaving_atoms)):
        raise ValueError("Duplicate atom indices found")

    return leaving_atoms


def remove_leaving_atoms(
    assembly: Assembly,
    chem_comp_dict: dict[str, ChemComp],
):
    mask = new_atom_mask(assembly)

    removed_atoms: list[int] = []
    for bond in assembly.connections:
        if bond.conn_type != "covale":
            continue

        if not bond.leaving_atom_count:
            continue

        for atom_idx in (bond.src_idx, bond.dst_idx):
            atom = assembly.atoms[atom_idx]
            if not _is_ligand(assembly.residues[atom.residue_id].chem_comp):
                continue

            ccd_data = chem_comp_dict.get(atom.comp_id, None)
            if ccd_data is None:
                continue

            residue = assembly.residues[atom.residue_id]
            removed_atoms += _find_leaving_atoms(assembly, residue, ccd_data)

    mask[removed_atoms] = True
    return mask


def filter_structure(
    assembly: Assembly,
    exptl_method: str,
    chem_comp_dict: dict[str, ChemComp],
    is_test: bool = False,
) -> NDArray[np.bool_]:
    mask = remove_min_resolved_chains(assembly)
    mask |= remove_waters(assembly)
    mask |= remove_all_unknown_chains(assembly)
    mask |= remove_clashing_chains(assembly)
    mask |= remove_non_ccd_atoms(assembly, chem_comp_dict)
    mask |= remove_large_ca_distance(assembly)
    mask |= remove_leaving_atoms(assembly, chem_comp_dict)
    if is_test:
        lig_mask = ligand_atom_mask(assembly)
        mask |= lig_mask & remove_crystallization_aids(assembly, exptl_method)
        mask |= lig_mask & remove_excluded_ligands(assembly)
    return mask


def filter_mmcif(
    assembly: Assembly,
    chem_comp_dict: dict[str, ChemComp],
    cutoff_date: dt.date = dt.date(2021, 9, 30),
    max_resolution: float = 9.0,
    max_chains: int = 300,
    is_test: bool = False,
) -> Assembly | None:
    if not (
        assembly.metadata.revision_date < cutoff_date
        and assembly.metadata.resolution < max_resolution
        and assembly.count_polymer_chains() <= max_chains
    ):
        return None

    mask = filter_structure(
        assembly,
        assembly.metadata.exptl_method,
        chem_comp_dict,
        is_test,
    )

    return assembly.filter(~mask)



import datetime as dt
import enum
import gzip
import logging
from collections.abc import Set
from dataclasses import InitVar, field
from pathlib import Path
from typing import TYPE_CHECKING

import networkx as nx
import numpy as np
from numpy.typing import NDArray
from pydantic import ConfigDict, TypeAdapter
from rdkit import Chem
from scipy.spatial import distance as D

from gmol.base import const
from .assembly import (
    Assembly,
    Chain,
    MolType,
    Residue,
    ResidueId,
    mmcif_assemblies,
)
from .filter import filter_mmcif
from .parse import (
    ChemComp,
    ChemCompAtom,
    ChemCompBond,
    load_mmcif_single,
)
from .smiles import (
    MmcifRefLigand,
    RefLigandInput,
    input_from_reference,
    reference_from_mmcif,
    unique_atom_id,
)

if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from pydantic.dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass
class PolymerConstants:
    restype_3_to_1: dict[str, str]
    restype_order_with_x: dict[str, int]

    backbone: set[str]
    modres_backbone_3: str

    bb_src_atom: str
    bb_src_leaving: list[str]
    bb_dst_atom: str
    bb_dst_leaving: list[str]

    max_res_atoms: int

    residue_atom_idxs: dict[str, dict[str, int]] = field(init=False)
    restype_modres_backbone: int = field(init=False)
    restype_x: int = field(init=False)

    _modres_backbone_1: InitVar[str]
    _res_to_atom: InitVar[dict[str, list[str]]]

    def __post_init__(
        self,
        _modres_backbone_1: str,
        _res_to_atom: dict[str, list[str]],
    ):
        # OXT etc. must also be considered as backbone atoms
        self.backbone.update(self.bb_src_leaving)
        self.backbone.update(self.bb_dst_leaving)

        self.residue_atom_idxs = {
            res: {
                atom_name: i
                for i, atom_name in enumerate(atoms)
                if atom_name != ""
            }
            for res, atoms in _res_to_atom.items()
        }
        self.restype_modres_backbone = self.restype_order_with_x[
            _modres_backbone_1
        ]
        self.restype_x = self.restype_order_with_x["X"]


chain_type_to_constants = {
    MolType.Protein: PolymerConstants(
        restype_3_to_1={**const.aa_restype_3to1, "MSE": "M"},
        restype_order_with_x=const.aa_restype_order_with_x,
        backbone={"N", "CA", "C", "O"},
        _modres_backbone_1="G",
        modres_backbone_3="GLY",
        bb_src_atom="C",
        bb_src_leaving=["OXT", "HXT"],
        bb_dst_atom="N",
        bb_dst_leaving=["H"],
        max_res_atoms=14,
        _res_to_atom=const.aa_restype_to_atom14,
    ),
    MolType.RNA: PolymerConstants(
        restype_3_to_1=const.rna_restype_3to1,
        restype_order_with_x=const.rna_restype_order_with_x,
        backbone=set(const.rna_backbone),
        _modres_backbone_1="X",
        modres_backbone_3="N",
        bb_src_atom="O3'",
        bb_src_leaving=["HO3'"],
        bb_dst_atom="P",
        bb_dst_leaving=["OP3", "HOP3"],
        max_res_atoms=24,
        _res_to_atom=const.rna_restype_to_atom24,
    ),
    MolType.DNA: PolymerConstants(
        restype_3_to_1=const.dna_restype_3to1,
        restype_order_with_x=const.dna_restype_order_with_x,
        backbone=set(const.dna_backbone),
        _modres_backbone_1="X",
        modres_backbone_3="DN",
        bb_src_atom="O3'",
        bb_src_leaving=["HO3'"],
        bb_dst_atom="P",
        bb_dst_leaving=["OP3", "HOP3"],
        max_res_atoms=24,
        _res_to_atom=const.dna_restype_to_atom24,
    ),
}


class BondOrder(enum.IntEnum):
    other = 0
    single = 1
    double = 2
    triple = 3
    quadruple = 4
    aromatic = 5

    @classmethod
    def from_rdkit_bond_type(cls, bond_type: Chem.BondType):
        mapping = {
            Chem.BondType.SINGLE: cls.single,
            Chem.BondType.DOUBLE: cls.double,
            Chem.BondType.TRIPLE: cls.triple,
            Chem.BondType.QUADRUPLE: cls.quadruple,
            Chem.BondType.AROMATIC: cls.aromatic,
        }
        return mapping.get(bond_type, cls.other)

    def to_rdkit_bond_type(self):
        mapping = {
            BondOrder.single: Chem.BondType.SINGLE,
            BondOrder.double: Chem.BondType.DOUBLE,
            BondOrder.triple: Chem.BondType.TRIPLE,
            BondOrder.quadruple: Chem.BondType.QUADRUPLE,
            BondOrder.aromatic: Chem.BondType.AROMATIC,
        }
        return mapping.get(self, Chem.BondType.UNSPECIFIED)

    @classmethod
    def from_mmcif_value_order(cls, value_order: int):
        mapping = {
            1: cls.single,
            2: cls.double,
            3: cls.triple,
            4: cls.quadruple,
        }
        return mapping.get(value_order, cls.other)


class BondClass(enum.IntEnum):
    unknown = 0
    covalent = 1
    disulfide = 2
    hydrogen = 3
    metal_coordinate = 4

    @classmethod
    def from_mmcif_conn_type(cls, conn_type: str):
        mapping = {
            "covale": cls.covalent,
            "disulf": cls.disulfide,
            "hydrog": cls.hydrogen,
            "metalc": cls.metal_coordinate,
        }
        return mapping.get(conn_type.lower(), cls.unknown)


@dataclass
class PolymerChain:
    mol_type: MolType

    entity_id: str
    chain_id: str

    # [N_res]
    # When set to X, modified residue (as a whole) is extracted as a ligand;
    # output corresponding to the residue must be ignored
    restype: NDArray[np.int8]
    # Raw chem_comp id for MSA generation; for modified residues
    chem_comp_raw: NDArray[np.str_]

    # [N_res, N_atom_type, 3]
    atom_coords: NDArray[np.float64]

    # [N_res, N_atom_type]
    # Temperature factor (B-factor) aligned with atom_coords. Missing values are NaN.
    atom_b_factors: NDArray[np.float64]

    # Below is for reference, not used in input
    # [N_res]
    residue_ids: list[ResidueId | None]

    __pydantic_config__ = ConfigDict(arbitrary_types_allowed=True)

    def __len__(self):
        return len(self.restype)


@dataclass
class NonPolymerLigand:
    entity_id: str
    chain_id: str

    smiles: str

    # [N_atom]
    atom_ids: NDArray[np.str_]
    # [N_atom, 3]
    atom_coords: NDArray[np.float64]

    # [N_atom]
    # Temperature factor (B-factor) aligned with atom_ids. Missing values are NaN.
    atom_b_factors: NDArray[np.float64]

    __pydantic_config__ = ConfigDict(arbitrary_types_allowed=True)


@dataclass(order=True, frozen=True)
class ExtraBondPartner:
    chain_id: str
    res_idx: int
    atom_id: str

    leaving_atoms: list[str] = field(default_factory=list, compare=False)


@dataclass
class ExtraBond:
    src: ExtraBondPartner
    dst: ExtraBondPartner

    bond_order: BondOrder
    bond_class: BondClass
    # E:-1. N/Undefined:0. Z:1  # Create a EZ stereo dataclass?
    bond_stereo_EZ: int = 0

    def __post_init__(self):
        if (
            self.bond_class != BondClass.covalent
            or self.bond_order != BondOrder.double
        ):
            assert self.bond_stereo_EZ == 0, (
                "No EZ Configuration except for double covalent bonds."
            )

    @staticmethod
    def stereo_from_mmcif(pdbx_stereo_config: str | None):
        return {"E": -1, "N": 0, None: 0, "Z": 1}.get(pdbx_stereo_config, 0)


@dataclass(frozen=True)
class _ModresFrom:
    chain: str
    res_idx: int
    atom_id: str


@dataclass
class _ModresTo:
    mod_chain: str
    res_idx: int


@dataclass
class Input:
    polymers: list[PolymerChain]
    ligands: list[NonPolymerLigand]
    extra_bonds: list[ExtraBond]

    release_date: dt.date
    resolution: float
    is_distillation: bool


def _build_atom_data_from_ref(
    assembly: Assembly,
    residues: list[Residue],
    ref_ligand: RefLigandInput,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    coords = np.full((len(ref_ligand.atom_ids), 3), np.nan, dtype=np.float64)
    b_factors = np.full((len(ref_ligand.atom_ids),), np.nan, dtype=np.float64)

    if len(residues) == 1:

        def resolve_atom_id(residue_id: ResidueId, atom_id: str):
            return atom_id
    else:

        def resolve_atom_id(residue_id: ResidueId, atom_id: str):
            return unique_atom_id(residue_id, atom_id)

    aid_to_index = {aid: i for i, aid in enumerate(ref_ligand.atom_ids)}
    for residue in residues:
        for atom in assembly.atoms_of_residue(residue):
            atom_id = resolve_atom_id(residue.residue_id, atom.atom_id)
            if atom_id not in aid_to_index:
                continue

            idx = aid_to_index[atom_id]
            coords[idx] = assembly.coords[atom.atom_idx]
            b_factors[idx] = atom.b_factor

    return coords, b_factors


def _update_polymer_residue_coords(
    coords: NDArray[np.float64],
    b_factors: NDArray[np.float64],
    assembly: Assembly,
    residue: Residue,
    polymer_consts: PolymerConstants,
    well_known_atoms: Set[str],
    comp_id: str | None = None,
):
    comp_id = comp_id or residue.chem_comp.id
    if comp_id == "MSE":
        # alias SE to SD
        atom_idxs = {
            ("SE" if aid == "SD" else aid): idx
            for aid, idx in polymer_consts.residue_atom_idxs["MET"].items()
        }
    else:
        atom_idxs = polymer_consts.residue_atom_idxs[comp_id]

    for atom in assembly.atoms_of_residue(residue):
        try:
            idx = atom_idxs[atom.atom_id]
            coords[idx] = assembly.coords[atom.atom_idx]
            b_factors[idx] = atom.b_factor
        except KeyError:
            if atom.atom_id not in well_known_atoms:
                logger.warning(
                    "Unknown atom %s in compound %s",
                    atom.atom_id,
                    comp_id,
                )

    if comp_id != "ARG":
        return

    # From AlphaFold2: swap NH1 and NH2 in ARG if NH2 is closer to CD than NH1
    i = atom_idxs["NH1"]
    j = atom_idxs["NH2"]
    cd_nh1_nh2 = coords[[atom_idxs["CD"], i, j]]
    if np.isnan(cd_nh1_nh2).any():
        return

    dsq_nh1 = D.sqeuclidean(cd_nh1_nh2[0], cd_nh1_nh2[1])
    dsq_nh2 = D.sqeuclidean(cd_nh1_nh2[0], cd_nh1_nh2[2])
    if dsq_nh1 > dsq_nh2:
        coords[[j, i]] = cd_nh1_nh2[1:]
        b_factors[[j, i]] = b_factors[[i, j]]


def _add_implicit_bb_bonds(
    polymer: PolymerChain,
    polymer_consts: PolymerConstants,
    mod_chain: str,
    res_idx: int,
):
    bonds: list[ExtraBond] = []
    if res_idx > 0:
        bonds.append(
            ExtraBond(
                src=ExtraBondPartner(
                    chain_id=polymer.chain_id,
                    res_idx=res_idx - 1,
                    atom_id=polymer_consts.bb_src_atom,
                    leaving_atoms=polymer_consts.bb_src_leaving,
                ),
                dst=ExtraBondPartner(
                    chain_id=mod_chain,
                    res_idx=0,
                    atom_id=polymer_consts.bb_dst_atom,
                    leaving_atoms=polymer_consts.bb_dst_leaving,
                ),
                bond_order=BondOrder.single,
                bond_class=BondClass.covalent,
            )
        )
    if res_idx < len(polymer) - 1:
        bonds.append(
            ExtraBond(
                src=ExtraBondPartner(
                    chain_id=mod_chain,
                    res_idx=0,
                    atom_id=polymer_consts.bb_src_atom,
                    leaving_atoms=polymer_consts.bb_src_leaving,
                ),
                dst=ExtraBondPartner(
                    chain_id=polymer.chain_id,
                    res_idx=res_idx + 1,
                    atom_id=polymer_consts.bb_dst_atom,
                    leaving_atoms=polymer_consts.bb_dst_leaving,
                ),
                bond_order=BondOrder.single,
                bond_class=BondClass.covalent,
            )
        )

    return bonds


def _modres_as_ligand(
    assembly: Assembly,
    polymer: PolymerChain,
    polymer_consts: PolymerConstants,
    mod_entity: str,
    mod_chain: str,
    resid: ResidueId | None,
    res_idx: int,
    chem_comp: ChemComp,
):
    ref_mmcif = MmcifRefLigand(chem_comp.atoms, chem_comp.bonds)
    ref_ligand = input_from_reference(ref_mmcif)
    crd, bfs = _build_atom_data_from_ref(
        assembly,
        [] if resid is None else [assembly.residues[resid]],
        ref_ligand,
    )

    ligand = NonPolymerLigand(
        entity_id=mod_entity,
        chain_id=mod_chain,
        smiles=ref_ligand.smiles,
        atom_ids=np.array(ref_ligand.atom_ids, dtype=np.str_),
        atom_coords=crd,
        atom_b_factors=bfs,
    )
    bonds = _add_implicit_bb_bonds(polymer, polymer_consts, mod_chain, res_idx)

    atom_map = {
        _ModresFrom(
            chain=polymer.chain_id,
            res_idx=res_idx,
            atom_id=atom_id,
        ): _ModresTo(mod_chain=mod_chain, res_idx=0)
        for atom_id in ref_ligand.atom_ids
    }

    return ligand, bonds, atom_map


def _split_modified_residue(
    chain_consts: PolymerConstants,
    chem_comp: ChemComp,
    chain: str,
    res_idx: int,
    mod_chain_prefix: str,
) -> tuple[list[ChemComp], list[ExtraBond]]:
    graph = nx.Graph()
    graph.add_nodes_from([(a.atom_id, {"data": a}) for a in chem_comp.atoms])
    graph.add_edges_from(
        [(b.atom_id_1, b.atom_id_2, {"data": b}) for b in chem_comp.bonds]
    )

    bb_all = chain_consts.backbone.copy()
    for aid in chain_consts.backbone:
        for n in graph[aid]:
            atom: ChemCompAtom = graph.nodes[n]["data"]
            if atom.type_symbol in "HD":
                bb_all.add(n)

    only_sc = graph.copy(as_view=False)
    only_sc.remove_nodes_from(bb_all)

    fragments: list[ChemComp] = []
    fragment_idxs: dict[str, int] = {}
    atom_ids: set[str]
    for i, atom_ids in enumerate(nx.connected_components(only_sc)):
        for atom_id in atom_ids:
            fragment_idxs[atom_id] = i

        frag_graph: nx.Graph = nx.induced_subgraph(only_sc, atom_ids)
        comp_atoms = [atom for _, atom in frag_graph.nodes(data="data")]
        comp_bonds = [bond for *_, bond in frag_graph.edges(data="data")]
        frag_cc = ChemComp(
            id=f"{chem_comp.id}-FRAG{i}",
            name=f"{chem_comp.name}-FRAG{i}",
            type=f"{chem_comp.type}, fragment {i}",
            formula=None,
            formula_weight=None,
            atoms=comp_atoms,
            bonds=comp_bonds,
        )
        fragments.append(frag_cc)

    inter_bonds: list[ExtraBond] = []
    hydrogen_counts = [0] * len(fragments)
    bond: ChemCompBond
    for src, dst, bond in nx.edge_boundary(graph, bb_all, data="data"):
        # TODO: handle value_order
        assert bond.value_order == 1

        fi = fragment_idxs[dst]

        inter_bonds.append(
            ExtraBond(
                src=ExtraBondPartner(
                    chain_id=chain,
                    res_idx=res_idx,
                    atom_id=src,
                ),
                dst=ExtraBondPartner(
                    chain_id=f"{mod_chain_prefix}-FRAG{fi}",
                    res_idx=0,
                    atom_id=dst,
                ),
                bond_order=BondOrder.from_mmcif_value_order(bond.value_order),
                bond_class=BondClass.covalent,
                bond_stereo_EZ=ExtraBond.stereo_from_mmcif(
                    bond.pdbx_stereo_config
                ),
            )
        )

        hname = f"{dst}-DUMMYH-{hydrogen_counts[fi]}"
        hydrogen_counts[fi] += 1

        fragment = fragments[fi]
        fragment.atoms.append(ChemCompAtom(atom_id=hname, type_symbol="H"))
        fragment.bonds.append(
            ChemCompBond(atom_id_1=dst, atom_id_2=hname, value_order=1)
        )

    return fragments, inter_bonds


def _build_sc_ligand(
    assembly: Assembly,
    resid: ResidueId | None,
    sc_frag: ChemComp,
    mod_entity: str,
    mod_chain: str,
) -> NonPolymerLigand:
    ref_mmcif = MmcifRefLigand(sc_frag.atoms, sc_frag.bonds)
    ref_ligand = input_from_reference(ref_mmcif)
    crd, bfs = _build_atom_data_from_ref(
        assembly,
        [] if resid is None else [assembly.residues[resid]],
        ref_ligand,
    )

    ligand = NonPolymerLigand(
        entity_id=mod_entity,
        chain_id=mod_chain,
        smiles=ref_ligand.smiles,
        atom_ids=np.array(ref_ligand.atom_ids, dtype=np.str_),
        atom_coords=crd,
        atom_b_factors=bfs,
    )
    return ligand


def _modres_split_bb_sc(
    assembly: Assembly,
    chain_consts: PolymerConstants,
    chain_id: str,
    resid: ResidueId | None,
    res_idx: int,
    chem_comp: ChemComp,
):
    mod_entity_prefix = f"X-MOD:{chem_comp.id}"
    mod_chain_prefix = f"{chain_id}-MOD{res_idx}"

    sidechain_frags, inter_bonds = _split_modified_residue(
        chain_consts,
        chem_comp,
        chain_id,
        res_idx,
        mod_chain_prefix,
    )
    sidechain_ligands = [
        _build_sc_ligand(
            assembly,
            resid,
            frag,
            f"{mod_entity_prefix}-FRAG{i}",
            f"{mod_chain_prefix}-FRAG{i}",
        )
        for i, frag in enumerate(sidechain_frags)
    ]

    atom_map = {
        _ModresFrom(
            chain=chain_id,
            res_idx=res_idx,
            atom_id=atom_id,
        ): _ModresTo(mod_chain=f"{mod_chain_prefix}-FRAG{i}", res_idx=0)
        for i, lig in enumerate(sidechain_ligands)
        for atom_id in lig.atom_ids
    }

    return sidechain_ligands, inter_bonds, atom_map


def process_polymer_chain(
    assembly: Assembly,
    chain: Chain,
    ccd: dict[str, ChemComp],
    split_modified: bool = False,
    well_known_atoms: Set[str] = frozenset({"OXT", "HXT"}),
):
    polymer_consts = chain_type_to_constants[chain.type]

    polymer = PolymerChain(
        mol_type=chain.type,
        entity_id=str(chain.entity_id),
        chain_id=chain.chain_id,
        restype=np.array(
            [
                polymer_consts.restype_order_with_x[
                    polymer_consts.restype_3_to_1.get(seq.comp_id, "X")
                ]
                for seq in chain.seqres
            ]
        ),
        chem_comp_raw=np.array(
            [seq.comp_id for seq in chain.seqres], dtype=np.str_
        ),
        atom_coords=np.full(
            (len(chain.seqres), polymer_consts.max_res_atoms, 3),
            np.nan,
            dtype=np.float64,
        ),
        atom_b_factors=np.full(
            (len(chain.seqres), polymer_consts.max_res_atoms),
            np.nan,
            dtype=np.float64,
        ),
        residue_ids=[seq.res_id for seq in chain.seqres],
    )

    non_polymers: list[NonPolymerLigand] = []
    extra_bonds: list[ExtraBond] = []
    modres_atom_map: dict[_ModresFrom, _ModresTo] = {}

    restype: int
    for i, (seq, restype) in enumerate(zip(chain.seqres, polymer.restype)):
        # Standard residues
        if restype != polymer_consts.restype_x:
            if seq.res_id is not None:
                _update_polymer_residue_coords(
                    polymer.atom_coords[i],
                    polymer.atom_b_factors[i],
                    assembly,
                    assembly.residues[seq.res_id],
                    polymer_consts,
                    well_known_atoms,
                )
            continue

        mod_entity = f"X-MOD:{seq.comp_id}"
        mod_chain = f"{chain.chain_id}-MOD{i}"

        # whole modified residue as ligand (no polymer)
        if not split_modified:
            lig, bonds, atom_map = _modres_as_ligand(
                assembly,
                polymer,
                polymer_consts,
                mod_entity,
                mod_chain,
                seq.res_id,
                i,
                ccd[seq.comp_id],
            )

            non_polymers.append(lig)
            extra_bonds.extend(bonds)
            modres_atom_map.update(atom_map)
            continue

        # split modified residue into backbone and sidechain "covalent ligands"
        polymer.restype[i] = polymer_consts.restype_modres_backbone

        if seq.res_id is not None:
            _update_polymer_residue_coords(
                polymer.atom_coords[i],
                polymer.atom_b_factors[i],
                assembly,
                assembly.residues[seq.res_id],
                polymer_consts,
                well_known_atoms,
                comp_id=polymer_consts.modres_backbone_3,
            )

        sidechain_ligands, inter_bonds_local, atom_map = _modres_split_bb_sc(
            assembly,
            polymer_consts,
            chain.chain_id,
            seq.res_id,
            i,
            ccd[seq.comp_id],
        )
        non_polymers.extend(sidechain_ligands)
        extra_bonds.extend(inter_bonds_local)
        modres_atom_map.update(atom_map)

    return polymer, non_polymers, extra_bonds, modres_atom_map


def process_ligand_chain(assembly: Assembly, chain: Chain):
    # TODO: handle ligand chain with multiple ligands
    ccs = [
        (residue.residue_id, residue.chem_comp)
        for residue in assembly.residues_of_chain(chain)
    ]
    lig = input_from_reference(reference_from_mmcif(ccs, chain.branches))
    residues = [assembly.residues[rid] for rid, _ in ccs]
    coords, b_factors = _build_atom_data_from_ref(assembly, residues, lig)

    return NonPolymerLigand(
        chain_id=chain.chain_id,
        entity_id=str(chain.entity_id),
        smiles=lig.smiles,
        atom_ids=np.array(lig.atom_ids, dtype=np.str_),
        atom_coords=coords,
        atom_b_factors=b_factors,
    )


def _resolve_partner(
    assembly: Assembly,
    modres_atom_map: dict[_ModresFrom, _ModresTo],
    atom_idx: int,
):
    atom = assembly.atoms[atom_idx]
    chain = assembly.chains[atom.chain_id]

    ptnr = ExtraBondPartner(
        chain_id=atom.chain_id,
        res_idx=chain.resseq[atom.residue_id],
        atom_id=atom.atom_id,
    )

    key = _ModresFrom(ptnr.chain_id, ptnr.res_idx, ptnr.atom_id)
    if (val := modres_atom_map.get(key, None)) is not None:
        ptnr = ExtraBondPartner(
            chain_id=val.mod_chain,
            res_idx=val.res_idx,
            atom_id=ptnr.atom_id,
        )
    return ptnr


def _is_implicit_backbone_bond(
    assembly: Assembly,
    src: ExtraBondPartner,
    dst: ExtraBondPartner,
):
    if src.chain_id != dst.chain_id:
        return False

    try:
        chain = assembly.chains[src.chain_id]
        polymer_consts = chain_type_to_constants[chain.type]
    except KeyError:
        return False

    src, dst = sorted([src, dst], key=lambda x: x.res_idx)
    return (
        src.res_idx + 1 == dst.res_idx
        and src.atom_id == polymer_consts.bb_src_atom
        and dst.atom_id == polymer_consts.bb_dst_atom
    )


def process_connections(
    assembly: Assembly,
    modres_atom_map: dict[_ModresFrom, _ModresTo],
) -> list[ExtraBond]:
    bonds: list[ExtraBond] = []

    for conn in assembly.connections:
        src = _resolve_partner(assembly, modres_atom_map, conn.src_idx)
        dst = _resolve_partner(assembly, modres_atom_map, conn.dst_idx)

        if _is_implicit_backbone_bond(assembly, src, dst):
            continue

        # TODO: handle leaving atoms
        bond = ExtraBond(
            src=src,
            dst=dst,
            bond_order=BondOrder.single,
            bond_class=BondClass.from_mmcif_conn_type(conn.conn_type),
        )
        bonds.append(bond)

    return bonds


def _bond_as_key(bond: ExtraBond):
    return tuple(
        sorted(
            [
                (bond.src.chain_id, bond.src.res_idx, bond.src.atom_id),
                (bond.dst.chain_id, bond.dst.res_idx, bond.dst.atom_id),
            ]
        )
    )


def build_input(
    assembly: Assembly,
    ccd: dict[str, ChemComp],
    split_modified: bool = False,
    well_known_atoms: Set[str] = frozenset({"OXT", "HXT"}),
) -> Input:
    # bonds for modified residues (backbone - sidechain)
    new_bonds_for_modres: list[ExtraBond] = []
    polymers_all: list[PolymerChain] = []
    non_polymers_all: list[NonPolymerLigand] = []
    modres_atom_map: dict[_ModresFrom, _ModresTo] = {}
    for chain in assembly.chains.values():
        if not chain.type.is_polymer:
            non_polymers_all.append(process_ligand_chain(assembly, chain))
            continue

        poly, nonpolys, modres_bonds, modres_map = process_polymer_chain(
            assembly,
            chain,
            ccd,
            split_modified,
            well_known_atoms,
        )
        polymers_all.append(poly)
        non_polymers_all.extend(nonpolys)
        new_bonds_for_modres.extend(modres_bonds)
        modres_atom_map.update(modres_map)

    explicit_bonds = process_connections(assembly, modres_atom_map)

    mmcif_bond_index = {_bond_as_key(bond): bond for bond in explicit_bonds}
    for bond in new_bonds_for_modres:
        if _bond_as_key(bond) in mmcif_bond_index:
            # See seoklab/gmol-base#9 for context.
            logger.debug("Duplicate bond: %s", bond)
        else:
            explicit_bonds.append(bond)

    return Input(
        polymers=polymers_all,
        ligands=non_polymers_all,
        extra_bonds=explicit_bonds,
        release_date=assembly.metadata.revision_date,
        resolution=assembly.metadata.resolution,
        is_distillation=False,
    )


def build_input_from_mmcif(
    mmcif_path: str,
    ccd_comp: dict[str, ChemComp],
) -> Input | None:
    metadata = load_mmcif_single(Path(mmcif_path))
    assemblies = mmcif_assemblies(metadata, ccd_comp)
    filtered_assembly = filter_mmcif(assemblies[0], ccd_comp)

    if filtered_assembly is not None:
        return build_input(filtered_assembly, ccd_comp, split_modified=False)

    return None


def load_ccd_dict_json(file_path: Path) -> dict[str, ChemComp]:
    with gzip.open(file_path) as f:
        ccd_comp = TypeAdapter(dict[str, ChemComp]).validate_json(f.read())
    return ccd_comp



import datetime as dt
import math
from pathlib import Path

import numpy as np
import pytest
from pydantic import TypeAdapter

from gmol.base.data.mmcif import (
    ChemComp,
    filter_mmcif,
    load_mmcif_single,
    mmcif_assemblies,
)
from gmol.base.data.mmcif.input import build_input


@pytest.fixture(scope="session")
def ccd_min(test_data: Path) -> dict[str, ChemComp]:
    return TypeAdapter(dict[str, ChemComp]).validate_json(
        test_data.joinpath("ccd", "components_min.json").read_bytes()
    )


def _build_input_for_pdb(
    pdb_id: str,
    test_data: Path,
    ccd: dict[str, ChemComp],
):
    mmcif = test_data / "mmcif" / f"{pdb_id}.cif"
    data = load_mmcif_single(mmcif)

    assemblies = mmcif_assemblies(data, ccd)
    assert len(assemblies) == 1

    filtered = filter_mmcif(
        assemblies[0], ccd, cutoff_date=dt.date(9999, 12, 31)
    )
    assert filtered is not None

    return build_input(filtered, ccd, split_modified=False)


def test_atom_site_bfactor_is_parsed(test_data: Path):
    mmcif = test_data / "mmcif" / "1ubq.cif"
    data = load_mmcif_single(mmcif)

    assert len(data.atom_site) > 0

    bvals = [a.b_iso_or_equiv for a in data.atom_site[:50]]
    assert any(v is not None for v in bvals), "B-factor should be present"

    non_null = [v for v in bvals if math.isfinite(v)]
    assert all(float(v) >= 0.0 for v in non_null)


def test_polymer_chain_bfactor_tensor(
    test_data: Path, ccd_min: dict[str, ChemComp]
):
    input_data = _build_input_for_pdb("4hf7", test_data, ccd_min)
    assert len(input_data.polymers) > 0

    poly = input_data.polymers[0]
    assert poly.atom_coords.shape[:2] == poly.atom_b_factors.shape

    # If a coordinate exists for an atom type, B-factor should also be set
    # (at least for some atoms; missing values are NaN)
    present = ~np.isnan(poly.atom_coords[..., 0])
    b_present = ~np.isnan(poly.atom_b_factors)
    assert int(np.sum(present & b_present)) > 0


def test_ligand_chain_bfactor_vector(
    test_data: Path, ccd_min: dict[str, ChemComp]
):
    input_data = _build_input_for_pdb("7qgw", test_data, ccd_min)
    assert len(input_data.ligands) > 0

    lig = next((l for l in input_data.ligands), None)
    assert lig is not None

    b_factors = lig.atom_b_factors
    assert b_factors is not None

    assert b_factors.shape == (len(lig.atom_ids),)
    assert int(np.sum(~np.isnan(b_factors))) > 0



from pathlib import Path

from gmol.base.data.mmcif import (
    ChemComp,
    ResidueId,
    SequenceToResidue,
    load_components,
    load_mmcif_single,
    mmcif_assemblies,
)


def test_pdb(test_data: Path):
    mmcif = test_data / "mmcif" / "1ubq.cif"
    data = load_mmcif_single(mmcif)

    assert data.entry_id == "1UBQ"


def test_ccd(test_data: Path):
    ccd = test_data / "ccd" / "components_stdres.cif"

    stdres = load_components(ccd)
    assert len(stdres) == 20
    assert list(stdres) == [
        "GLY",
        "ALA",
        "VAL",
        "LEU",
        "ILE",
        "THR",
        "SER",
        "MET",
        "CYS",
        "PRO",
        "PHE",
        "TYR",
        "TRP",
        "HIS",
        "LYS",
        "ARG",
        "ASP",
        "GLU",
        "ASN",
        "GLN",
    ]


def test_load_inconsistent_seq(
    test_data: Path,
    ccd_components: dict[str, ChemComp],
):
    mmcif = test_data / "mmcif" / "3m8z.cif"
    data = load_mmcif_single(mmcif)

    ccd_tmp = ccd_components.copy()
    ccd_tmp["TPO"] = ChemComp.model_validate(
        {
            "id": "TPO",
            "name": "PHOSPHOTHREONINE",
            "type": "L-PEPTIDE LINKING",
            "formula": "C4 H10 N O6 P",
            "formula_weight": 199.099,
            "atoms": [
                {
                    "atom_id": "N",
                    "type_symbol": "N",
                },
                {
                    "atom_id": "CA",
                    "type_symbol": "C",
                },
                {
                    "atom_id": "CB",
                    "type_symbol": "C",
                },
                {
                    "atom_id": "CG2",
                    "type_symbol": "C",
                },
                {
                    "atom_id": "OG1",
                    "type_symbol": "O",
                },
                {
                    "atom_id": "P",
                    "type_symbol": "P",
                },
                {
                    "atom_id": "O1P",
                    "type_symbol": "O",
                },
                {
                    "atom_id": "O2P",
                    "type_symbol": "O",
                },
                {
                    "atom_id": "O3P",
                    "type_symbol": "O",
                },
                {
                    "atom_id": "C",
                    "type_symbol": "C",
                },
                {
                    "atom_id": "O",
                    "type_symbol": "O",
                },
                {
                    "atom_id": "OXT",
                    "type_symbol": "O",
                    "pdbx_leaving_atom_flag": True,
                },
            ],
            "bonds": [
                {
                    "atom_id_1": "N",
                    "atom_id_2": "CA",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "CA",
                    "atom_id_2": "CB",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "CA",
                    "atom_id_2": "C",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "CB",
                    "atom_id_2": "CG2",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "CB",
                    "atom_id_2": "OG1",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "OG1",
                    "atom_id_2": "P",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "P",
                    "atom_id_2": "O1P",
                    "value_order": 2,
                },
                {
                    "atom_id_1": "P",
                    "atom_id_2": "O2P",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "P",
                    "atom_id_2": "O3P",
                    "value_order": 1,
                },
                {
                    "atom_id_1": "C",
                    "atom_id_2": "O",
                    "value_order": 2,
                },
                {
                    "atom_id_1": "C",
                    "atom_id_2": "OXT",
                    "value_order": 1,
                },
            ],
        }
    )

    assemblies = mmcif_assemblies(data, ccd_tmp)
    assert len(assemblies) == 3
    asm = assemblies[0]

    res_id = ResidueId("A", 85)

    for atom in asm.atoms_of_residue(asm.residues[res_id]):
        assert atom.comp_id == "THR"

    het_chains: list[SequenceToResidue] = [
        s2r for s2r in asm.chains["A"].seqres if s2r.res_id == res_id
    ]
    assert len(het_chains) == 1
    assert het_chains[0].comp_id == "THR"



import numpy as np
import pytest

from gmol.base.data.mmcif import Assembly, ChemComp, filter_mmcif


def test_filter_mmcif(
    ccd_components: dict[str, ChemComp],
    sample_assembly: Assembly,
):
    result = filter_mmcif(sample_assembly, ccd_components)
    if result is None:
        pytest.skip("target did not pass pre-filter")

    assert not np.isnan(result.coords).any()

    assert "A" in result.chains  # chain A is protein
    assert "B" not in result.chains  # chain B is water molecules



import enum
import itertools
import math
from collections import defaultdict
from collections.abc import Iterable
from copy import deepcopy
from dataclasses import dataclass, field, fields
from functools import cached_property
from string import ascii_lowercase, ascii_uppercase, digits
from typing import ClassVar, Protocol

import numpy as np
from numpy.typing import NDArray
from pydantic import (
    SerializerFunctionWrapHandler,
    TypeAdapter,
    field_serializer,
    field_validator,
    model_serializer,
    model_validator,
)

from gmol.base.const import (
    aa_restype_3to1,
    dna_restype_3to1,
    modres,
    rna_restype_3to1,
)
from gmol.base.types import LooseModel
from .parse import (
    AtomSite,
    BioAssembly,
    BranchLink,
    BranchLinkPartner,
    ChemComp,
    Entity,
    Mmcif,
    PdbMetadata,
    Scheme,
    StructConn,
    SymOp,
)
from .write import mmcif_bond_order, mmcif_bool, mmcif_write_block

__all__ = [
    "Assembly",
    "AssemblyAtom",
    "AssemblyConnection",
    "Chain",
    "MolType",
    "Residue",
    "ResidueId",
    "SequenceToResidue",
    "Transformation",
    "expand_ops",
    "mmcif_assemblies",
    "mmcif_chain_types",
    "polymer_mol_type",
    "residue_mol_type",
]


class MolType(enum.IntEnum):
    Protein = 1
    RNA = 2
    DNA = 3
    Ligand = 4

    @property
    def is_polymer(self) -> bool:
        return self.value < MolType.Ligand


@dataclass(frozen=True, order=True)
class ResidueId:
    chain_id: str
    seq_id: int
    ins_code: str = ""

    def __str__(self):
        return f"{self.chain_id}.{self.seq_id}{self.ins_code}"

    @classmethod
    def from_scheme(cls, scheme: Scheme):
        if scheme.pdb_seq_num is None:
            return None

        return cls(
            scheme.asym_id,
            scheme.pdb_seq_num,
            scheme.pdb_ins_code or "",
        )

    def with_chain_suffix(self, suffix: str):
        return ResidueId(self.chain_id + suffix, self.seq_id, self.ins_code)


@dataclass(order=True)
class Residue:
    residue_id: ResidueId
    chem_comp: ChemComp = field(compare=False)

    atom_idxs: list[int] = field(default_factory=list, compare=False)

    def new_residue(self, atom_idxs: list[int]):
        return Residue(self.residue_id, self.chem_comp, atom_idxs)

    def with_updates(self, offset: int, chain_suffix: str):
        return Residue(
            self.residue_id.with_chain_suffix(chain_suffix),
            self.chem_comp,
            [idx + offset for idx in self.atom_idxs],
        )


@dataclass
class SequenceToResidue:
    comp_id: str
    seq_id: int

    res_id: ResidueId | None

    @classmethod
    def from_scheme(cls, scheme: Scheme):
        return cls(scheme.mon_id, scheme.seq_id, ResidueId.from_scheme(scheme))

    def with_chain_suffix(self, suffix: str):
        return SequenceToResidue(
            self.comp_id,
            self.seq_id,
            self.res_id.with_chain_suffix(suffix) if self.res_id else None,
        )


@dataclass(order=True, frozen=True)
class BranchPartner:
    seq_id: int
    atom_id: str
    leaving_atom_id: str

    comp_id: str = field(compare=False)
    res_id: ResidueId

    @classmethod
    def from_branch_scheme(cls, scheme: Scheme, br_info: BranchLinkPartner):
        res_id = ResidueId.from_scheme(scheme)
        assert res_id is not None

        return cls(
            scheme.seq_id,
            br_info.atom_id,
            br_info.leaving_atom_id,
            scheme.mon_id,
            res_id,
        )

    def with_chain_suffix(self, suffix: str):
        return BranchPartner(
            self.seq_id,
            self.atom_id,
            self.leaving_atom_id,
            self.comp_id,
            self.res_id.with_chain_suffix(suffix),
        )

    def to_scheme(self, entity_id: int) -> Scheme:
        return Scheme(
            asym_id=self.res_id.chain_id,
            entity_id=entity_id,
            mon_id=self.comp_id,
            seq_id=self.seq_id,
            pdb_seq_num=self.res_id.seq_id,
            pdb_ins_code=self.res_id.ins_code or None,
        )


@dataclass(order=True, frozen=True)
class Branch:
    ptnr1: BranchPartner
    ptnr2: BranchPartner
    order: int = field(compare=False)

    def with_chain_suffix(self, suffix: str):
        return Branch(
            self.ptnr1.with_chain_suffix(suffix),
            self.ptnr2.with_chain_suffix(suffix),
            self.order,
        )


@dataclass(order=True)
class Chain:
    chain_id: str

    entity_id: int = field(compare=False)
    type: MolType = field(compare=False)

    seqres: list[SequenceToResidue] = field(
        default_factory=list, compare=False
    )

    branches: list[Branch] = field(default_factory=list, compare=False)

    residue_ids: list[ResidueId] = field(default_factory=list, compare=False)
    atom_idxs: list[int] = field(default_factory=list, compare=False)

    @cached_property
    def resseq(self) -> dict[ResidueId, int]:
        return {
            seq.res_id: i
            for i, seq in enumerate(self.seqres)
            if seq.res_id is not None
        }

    def new_chain(self, residue_ids: list[ResidueId], atom_idxs: list[int]):
        chain = Chain(
            self.chain_id,
            self.entity_id,
            self.type,
            deepcopy(self.seqres),
            deepcopy(self.branches),
            residue_ids,
            atom_idxs,
        )
        chain.sync_mappings()
        return chain

    def with_updates(self, offset: int, suffix: str):
        return Chain(
            self.chain_id + suffix,
            self.entity_id,
            self.type,
            [seqres.with_chain_suffix(suffix) for seqres in self.seqres],
            [branch.with_chain_suffix(suffix) for branch in self.branches],
            [rid.with_chain_suffix(suffix) for rid in self.residue_ids],
            [idx + offset for idx in self.atom_idxs],
        )

    def sync_mappings(self):
        res_ids = set(self.residue_ids)

        for seqres in self.seqres:
            if seqres.res_id is not None and seqres.res_id not in res_ids:
                seqres.res_id = None

        self.branches = [
            branch
            for branch in self.branches
            if branch.ptnr1.res_id in res_ids
            and branch.ptnr2.res_id in res_ids
        ]


@dataclass
class AssemblyAtom:
    atom_idx: int
    residue_id: ResidueId

    type_symbol: str
    atom_id: str
    comp_id: str

    occupancy: float
    b_factor: float

    @property
    def chain_id(self) -> str:
        return self.residue_id.chain_id

    def new_atom(self, idx: int):
        return AssemblyAtom(
            atom_idx=idx,
            residue_id=self.residue_id,
            type_symbol=self.type_symbol,
            atom_id=self.atom_id,
            comp_id=self.comp_id,
            occupancy=self.occupancy,
            b_factor=self.b_factor,
        )

    def with_updates(self, idx: int, chain_suffix: str):
        return AssemblyAtom(
            atom_idx=idx,
            residue_id=self.residue_id.with_chain_suffix(chain_suffix),
            type_symbol=self.type_symbol,
            atom_id=self.atom_id,
            comp_id=self.comp_id,
            occupancy=self.occupancy,
            b_factor=self.b_factor,
        )


class _IndexMapper(Protocol):
    def __getitem__(self, idx: int) -> int: ...


@dataclass
class AssemblyConnection:
    src_idx: int
    dst_idx: int

    conn_type: str
    leaving_atom_count: int

    def __post_init__(self):
        self.src_idx = int(self.src_idx)
        self.dst_idx = int(self.dst_idx)
        self.leaving_atom_count = int(self.leaving_atom_count)

    def remap(self, idx_map: _IndexMapper):
        return AssemblyConnection(
            src_idx=idx_map[self.src_idx],
            dst_idx=idx_map[self.dst_idx],
            conn_type=self.conn_type,
            leaving_atom_count=self.leaving_atom_count,
        )


class Transformation:
    ident: ClassVar["Transformation"] = None  # type: ignore[assignment]

    def __init__(
        self, matrix: NDArray[np.float64], vector: NDArray[np.float64]
    ):
        self.matrix = matrix
        self.vector = vector

    @classmethod
    def from_symop(cls, op: SymOp):
        return cls(op.matrix, op.vector)

    def __add__(self, other: "Transformation") -> "Transformation":
        return Transformation(
            matrix=self.matrix @ other.matrix,
            vector=self.matrix @ other.vector + self.vector,
        )

    def __matmul__(self, coords: NDArray[np.float64]) -> NDArray[np.float64]:
        return (self.matrix @ coords.T + self.vector[:, None]).T


Transformation.ident = Transformation(
    matrix=np.eye(3, dtype=np.float64), vector=np.zeros(3, dtype=np.float64)
)


class _OffsetMapper:
    def __init__(self, offset: int):
        self.offset = offset

    def __getitem__(self, idx: int) -> int:
        return idx + self.offset


@dataclass
class _PdbAtom:
    record: str
    res_id: ResidueId
    res_name: str
    name: str
    coords: NDArray[np.float64]
    element: str
    occupancy: float
    b_factor: float

    def __post_init__(self):
        assert len(self.record) == 6
        assert len(self.res_id.chain_id) == 1
        assert -999 <= self.res_id.seq_id <= 9999
        assert len(self.res_id.ins_code) <= 1
        assert len(self.res_name) <= 3
        assert len(self.element) <= 2

    def to_pdb_line(self, serial: int):
        b = self.b_factor if math.isfinite(self.b_factor) else 0.0
        return (
            # 1-21
            f"{self.record}{serial:>5} {self.name} {self.res_name:>3} "
            # 22-27
            f"{self.res_id.chain_id}{self.res_id.seq_id:4d}{self.res_id.ins_code:1}"
            # 28-54
            f"   {self.coords[0]:>8.3f}{self.coords[1]:>8.3f}{self.coords[2]:>8.3f}"
            #                       6         7
            #                       12345678901234567
            f"{self.occupancy:>6.2f}{b:>6.2f}          {self.element.upper():>2}  "
        )


class Assembly(LooseModel):
    metadata: PdbMetadata
    coords: NDArray[np.float64]
    atoms: list[AssemblyAtom]
    residues: dict[ResidueId, Residue]
    chains: dict[str, Chain]
    entities: dict[int, Entity]
    connections: list[AssemblyConnection]

    @cached_property
    def comp_ids(self) -> NDArray[np.str_]:
        return np.array([atom.comp_id for atom in self.atoms], dtype=np.str_)

    @cached_property
    def chem_comps(self) -> dict[str, ChemComp]:
        return {
            res.chem_comp.id: res.chem_comp for res in self.residues.values()
        }

    @field_serializer("coords", mode="plain")
    @staticmethod
    def _serialize_coords(coords: NDArray[np.float64]):
        return coords.tolist()

    @field_validator("coords", mode="before")
    @staticmethod
    def _validate_coords(data):
        return np.array(data, dtype=np.float64)

    @field_serializer("residues", mode="plain")
    @staticmethod
    def _serialize_residues(residues: dict[ResidueId, Residue]):
        res = [
            {
                f.name: getattr(residue, f.name)
                for f in fields(residue)
                if f.name != "chem_comp"
            }
            | {"comp_id": residue.chem_comp.id}
            for residue in residues.values()
        ]
        return res

    @model_serializer(mode="wrap")
    def _gather_chem_comps(self, handler: SerializerFunctionWrapHandler):
        ccd = self.chem_comps
        data = handler(self)
        data["chem_comps"] = ccd
        return data

    @field_validator("residues", mode="before")
    @staticmethod
    def _validate_residues(residues):
        if isinstance(residues, dict):
            return residues

        residues = {ResidueId(**res["residue_id"]): res for res in residues}
        return residues

    @model_validator(mode="before")
    @staticmethod
    def _scatter_chem_comps(data):
        try:
            ccd = data.pop("chem_comps")
        except Exception:
            return data

        ccd = TypeAdapter(dict[str, ChemComp]).validate_python(ccd)

        for res in data["residues"]:
            res["chem_comp"] = ccd[res.pop("comp_id")]

        return data

    def count_polymer_chains(self) -> int:
        return len(self.chains) - sum(
            chain.type == MolType.Ligand for chain in self.chains.values()
        )

    def chains_of_type(self, tp: MolType):
        for chain in self.chains.values():
            if chain.type == tp:
                yield chain

    def has_chain_type(self, tp: MolType) -> bool:
        return any(chain.type == tp for chain in self.chains.values())

    def residues_of_chain(self, chain: Chain):
        for rid in chain.residue_ids:
            yield self.residues[rid]

    def atoms_of_chain(self, chain: Chain):
        for aid in chain.atom_idxs:
            yield self.atoms[aid]

    def atoms_of_residue(self, residue: Residue):
        for aid in residue.atom_idxs:
            yield self.atoms[aid]

    def filter(self, selected: NDArray[np.bool_]) -> "Assembly":
        idx_map = np.cumsum(selected.astype(np.int64)) - 1

        coords = self.coords[selected]

        atoms = [
            atom.new_atom(i)
            for i, atom in enumerate(itertools.compress(self.atoms, selected))
        ]

        residues: dict[ResidueId, Residue] = {}
        for rid, residue in self.residues.items():
            atom_idxs = np.array(residue.atom_idxs)
            atom_idxs = idx_map[atom_idxs[selected[atom_idxs]]].tolist()
            if not atom_idxs:
                continue

            residues[rid] = residue.new_residue(atom_idxs)

        chains: dict[str, Chain] = {}
        for cid, chain in self.chains.items():
            atom_idxs = np.array(chain.atom_idxs)
            atom_idxs = idx_map[atom_idxs[selected[atom_idxs]]].tolist()
            if not atom_idxs:
                continue

            chains[cid] = chain.new_chain(
                [rid for rid in chain.residue_ids if rid in residues],
                atom_idxs,
            )

        if self.connections:
            conn_idxs = np.array(
                [(c.src_idx, c.dst_idx) for c in self.connections],
            )
            selected_conns: Iterable[int] = (
                selected[conn_idxs].all(axis=1).nonzero()[0]
            )
            connections = [
                self.connections[i].remap(idx_map) for i in selected_conns
            ]
        else:
            connections = []

        eids = {chain.entity_id for chain in chains.values()}
        entities = {eid: self.entities[eid] for eid in eids}

        return Assembly(
            metadata=self.metadata,
            coords=coords,
            atoms=atoms,
            residues=residues,
            chains=chains,
            entities=entities,
            connections=connections,
        )

    def filter_chains(self, chain_ids: list[str]):
        atom_chain_ids = np.array(
            [atom.chain_id for atom in self.atoms], dtype=np.str_
        )
        selected = np.isin(atom_chain_ids, chain_ids)
        return self.filter(selected)

    def transform(self, xform: Transformation) -> "Assembly":
        return self.model_copy(update={"coords": xform @ self.coords})

    def apply(self, operations: list[Transformation]) -> "Assembly":
        return Assembly.join([self.transform(op) for op in operations])

    @classmethod
    def join(cls, assemblies: list["Assembly"]):
        if len(assemblies) == 1:
            return assemblies[0]

        coords = np.concatenate([assembly.coords for assembly in assemblies])

        atoms: list[AssemblyAtom] = []
        residues: dict[ResidueId, Residue] = {}
        chains: dict[str, Chain] = {}
        connections: list[AssemblyConnection] = []

        for i, assembly in enumerate(assemblies, 1):
            suffix = f"_{i}"
            offset = len(atoms)

            atoms.extend(
                atom.with_updates(j, suffix)
                for j, atom in enumerate(assembly.atoms, offset)
            )

            for residue in assembly.residues.values():
                residue = residue.with_updates(offset, suffix)
                residues[residue.residue_id] = residue

            for chain in assembly.chains.values():
                chain = chain.with_updates(offset, suffix)
                chains[chain.chain_id] = chain

            connections.extend(
                [
                    conn.remap(_OffsetMapper(offset))
                    for conn in assembly.connections
                ]
            )

        entities = dict(
            itertools.chain.from_iterable(
                asm.entities.items() for asm in assemblies
            )
        )

        return cls(
            metadata=assemblies[0].metadata,
            coords=coords,
            atoms=atoms,
            residues=residues,
            chains=chains,
            entities=entities,
            connections=connections,
        )

    def to_mmcif(self, name: str, write_schemes: bool = True) -> str:
        merged_branches: dict[int, set[Branch]] = defaultdict(set)
        branch_scheme: dict[str, set[Scheme]] = defaultdict(set)
        nonpoly_scheme: dict[str, set[Scheme]] = defaultdict(set)
        for chain in self.chains.values():
            merged_branches[chain.entity_id].update(chain.branches)

            br_schemes = branch_scheme[chain.chain_id]
            br_schemes.update(
                {
                    branch.ptnr1.to_scheme(chain.entity_id)
                    for branch in chain.branches
                },
                {
                    branch.ptnr2.to_scheme(chain.entity_id)
                    for branch in chain.branches
                },
            )

            if chain.type != MolType.Ligand:
                continue

            np_schemes = nonpoly_scheme[chain.chain_id]
            for seqres in chain.seqres:
                if seqres.res_id is None:
                    continue

                res = self.residues[seqres.res_id]
                scheme = Scheme(
                    asym_id=chain.chain_id,
                    entity_id=chain.entity_id,
                    mon_id=res.chem_comp.id,
                    seq_id=seqres.seq_id,
                    pdb_seq_num=res.residue_id.seq_id,
                    pdb_ins_code=res.residue_id.ins_code or None,
                )
                if scheme not in br_schemes:
                    np_schemes.add(scheme)

        label_seqs: dict[ResidueId, int | str] = {
            seqres.res_id: seqres.seq_id if chain.type.is_polymer else "."
            for chain in self.chains.values()
            for seqres in chain.seqres
            if seqres.res_id is not None
        }

        m = self.metadata
        mmcif_content = (
            f"data_{name}"
            + mmcif_write_block("entry", ["id"], [(m.entry_id,)])
            + mmcif_write_block("exptl", ["method"], [(m.exptl_method,)])
            + mmcif_write_block(
                "struct_keywords",
                ["pdbx_keywords"],
                [(m.pdbx_keywords,)],
            )
            + mmcif_write_block(
                "pdbx_audit_revision_history",
                ["ordinal", "revision_date"],
                [(1, m.revision_date.isoformat())],
            )
            + mmcif_write_block(
                "refine",
                ["ls_d_res_high"],
                [(f"{m.resolution:.2f}",)],
            )
        )

        mmcif_content += (
            mmcif_write_block(
                "entity",
                ["id", "type", "pdbx_description"],
                [
                    (entity.id, entity.type, entity.pdbx_description)
                    for entity in self.entities.values()
                ],
            )
            + mmcif_write_block(
                "struct_asym",
                ["id", "entity_id"],
                [(cid, chain.entity_id) for cid, chain in self.chains.items()],
            )
            + mmcif_write_block(
                "atom_site",
                [
                    "id",
                    "type_symbol",
                    "group_PDB",
                    "label_atom_id",
                    "label_alt_id",
                    "label_comp_id",
                    "label_asym_id",
                    "label_seq_id",
                    "auth_seq_id",
                    "auth_comp_id",
                    "auth_asym_id",
                    "pdbx_PDB_ins_code",
                    "pdbx_PDB_model_num",
                    "Cartn_x",
                    "Cartn_y",
                    "Cartn_z",
                    "occupancy",
                ],
                [
                    (
                        atom.atom_idx,
                        atom.type_symbol,
                        "ATOM",
                        atom.atom_id,
                        ".",
                        atom.comp_id,
                        atom.chain_id,
                        label_seqs[atom.residue_id],
                        atom.residue_id.seq_id,
                        atom.comp_id,
                        atom.chain_id,
                        atom.residue_id.ins_code or ".",
                        1,
                        f"{crd[0]:.3f}",
                        f"{crd[1]:.3f}",
                        f"{crd[2]:.3f}",
                        f"{atom.occupancy:.2f}",
                    )
                    for atom, crd in zip(self.atoms, self.coords)
                ],
            )
            + mmcif_write_block(
                "chem_comp",
                ["id", "type", "mon_nstd_flag"],
                [
                    (
                        comp.id,
                        comp.type,
                        "y" if comp.mon_nstd_flag else ".",
                    )
                    for comp in self.chem_comps.values()
                ],
            )
            + mmcif_write_block(
                "chem_comp_atom",
                [
                    "comp_id",
                    "atom_id",
                    "type_symbol",
                    "pdbx_aromatic_flag",
                    "pdbx_stereo_config",
                ],
                [
                    (
                        comp_id,
                        atom.atom_id,
                        atom.type_symbol,
                        mmcif_bool(atom.pdbx_aromatic_flag),
                        atom.pdbx_stereo_config or "N",
                    )
                    for comp_id, comp in self.chem_comps.items()
                    for atom in comp.atoms
                ],
            )
            + mmcif_write_block(
                "chem_comp_bond",
                [
                    "comp_id",
                    "atom_id_1",
                    "atom_id_2",
                    "value_order",
                    "pdbx_aromatic_flag",
                    "pdbx_stereo_config",
                ],
                [
                    (
                        comp_id,
                        bond.atom_id_1,
                        bond.atom_id_2,
                        mmcif_bond_order(bond.value_order),
                        mmcif_bool(bond.pdbx_aromatic_flag),
                        bond.pdbx_stereo_config or "N",
                    )
                    for comp_id, comp in self.chem_comps.items()
                    for bond in comp.bonds
                ],
            )
            + mmcif_write_block(
                "struct_conn",
                [
                    "id",
                    "conn_type_id",
                    "pdbx_leaving_atom_flag",
                    "pdbx_dist_value",
                    "ptnr1_label_atom_id",
                    "ptnr1_label_comp_id",
                    "ptnr1_label_asym_id",
                    "ptnr1_label_seq_id",
                    "ptnr1_auth_seq_id",
                    "ptnr1_auth_comp_id",
                    "ptnr1_auth_asym_id",
                    "pdbx_ptnr1_PDB_ins_code",
                    "ptnr1_symmetry",
                    "ptnr2_label_atom_id",
                    "ptnr2_label_comp_id",
                    "ptnr2_label_asym_id",
                    "ptnr2_label_seq_id",
                    "ptnr2_auth_seq_id",
                    "ptnr2_auth_comp_id",
                    "ptnr2_auth_asym_id",
                    "pdbx_ptnr2_PDB_ins_code",
                    "ptnr2_symmetry",
                ],
                [
                    (
                        str(i),
                        conn.conn_type,
                        {0: "none", 1: "one", 2: "both"}[
                            conn.leaving_atom_count
                        ],
                        ".",
                        (ptnr1 := self.atoms[conn.src_idx]).atom_id,
                        ptnr1.comp_id,
                        ptnr1.chain_id,
                        label_seqs[ptnr1.residue_id],
                        ptnr1.residue_id.seq_id,
                        ptnr1.comp_id,
                        ptnr1.chain_id,
                        ptnr1.residue_id.ins_code or ".",
                        "1_555",  # identity placeholder
                        (ptnr2 := self.atoms[conn.dst_idx]).atom_id,
                        ptnr2.comp_id,
                        ptnr2.chain_id,
                        label_seqs[ptnr2.residue_id],
                        ptnr2.residue_id.seq_id,
                        ptnr2.comp_id,
                        ptnr2.chain_id,
                        ptnr2.residue_id.ins_code or ".",
                        "1_555",  # identity placeholder
                    )
                    for i, conn in enumerate(self.connections, start=1)
                ],
            )
            + mmcif_write_block(
                "pdbx_entity_branch_link",
                [
                    "entity_id",
                    "entity_branch_list_num_1",
                    "comp_id_1",
                    "atom_id_1",
                    "leaving_atom_id_1",
                    "entity_branch_list_num_2",
                    "comp_id_2",
                    "atom_id_2",
                    "leaving_atom_id_2",
                    "value_order",
                ],
                [
                    (
                        entity_id,
                        branch.ptnr1.seq_id,
                        branch.ptnr1.comp_id,
                        branch.ptnr1.atom_id,
                        branch.ptnr1.leaving_atom_id,
                        branch.ptnr2.seq_id,
                        branch.ptnr2.comp_id,
                        branch.ptnr2.atom_id,
                        branch.ptnr2.leaving_atom_id,
                        mmcif_bond_order(branch.order),
                    )
                    for entity_id, branches in merged_branches.items()
                    for branch in branches
                ],
            )
        )

        if write_schemes:
            mmcif_content += (
                mmcif_write_block(
                    "pdbx_poly_seq_scheme",
                    [
                        "asym_id",
                        "entity_id",
                        "seq_id",
                        "mon_id",
                        "pdb_seq_num",
                        "pdb_strand_id",
                        "pdb_ins_code",
                    ],
                    [
                        (
                            cid,
                            chain.entity_id,
                            seqres.seq_id,
                            seqres.comp_id,
                            seqres.res_id.seq_id if seqres.res_id else "?",
                            cid,
                            (seqres.res_id.ins_code or ".")
                            if seqres.res_id
                            else "?",
                        )
                        for cid, chain in self.chains.items()
                        if chain.type.is_polymer
                        for seqres in chain.seqres
                    ],
                )
                + mmcif_write_block(
                    "pdbx_branch_scheme",
                    [
                        "asym_id",
                        "entity_id",
                        "num",
                        "mon_id",
                        "pdb_seq_num",
                        "pdb_ins_code",
                    ],
                    [
                        (
                            scheme.asym_id,
                            scheme.entity_id,
                            scheme.seq_id,
                            scheme.mon_id,
                            scheme.pdb_seq_num,
                            scheme.pdb_ins_code or ".",
                        )
                        for schemes in branch_scheme.values()
                        for scheme in sorted(schemes)
                        if scheme.pdb_seq_num is not None
                    ],
                )
                + mmcif_write_block(
                    "pdbx_nonpoly_scheme",
                    [
                        "asym_id",
                        "entity_id",
                        "mon_id",
                        "ndb_seq_num",
                        "pdb_seq_num",
                        "pdb_ins_code",
                    ],
                    [
                        (
                            scheme.asym_id,
                            scheme.entity_id,
                            scheme.mon_id,
                            scheme.seq_id,
                            scheme.pdb_seq_num,
                            scheme.pdb_ins_code or ".",
                        )
                        for schemes in nonpoly_scheme.values()
                        for scheme in sorted(schemes)
                        if scheme.pdb_seq_num is not None
                    ],
                )
            )

        return mmcif_content + "\n#\n"

    def to_mmcif_chain(self, name: str, cid: str, write_schemes: bool = True):
        chain = self.chains[cid]

        merged_branches = set(chain.branches)
        branch_scheme = {
            branch.ptnr1.to_scheme(chain.entity_id)
            for branch in chain.branches
        } | {
            branch.ptnr2.to_scheme(chain.entity_id)
            for branch in chain.branches
        }
        nonpoly_scheme: set[Scheme] = set()
        if chain.type == MolType.Ligand:
            for seqres in chain.seqres:
                if seqres.res_id is None:
                    continue

                res = self.residues[seqres.res_id]
                scheme = Scheme(
                    asym_id=cid,
                    entity_id=chain.entity_id,
                    mon_id=res.chem_comp.id,
                    seq_id=seqres.seq_id,
                    pdb_seq_num=res.residue_id.seq_id,
                    pdb_ins_code=res.residue_id.ins_code or None,
                )
                if scheme not in branch_scheme:
                    nonpoly_scheme.add(scheme)

        label_seqs: dict[ResidueId, int | str] = {
            seqres.res_id: seqres.seq_id if chain.type.is_polymer else "."
            for seqres in chain.seqres
            if seqres.res_id is not None
        }
        entity = self.entities[chain.entity_id]

        ccid_chain = {
            self.residues[rid].chem_comp.id for rid in chain.residue_ids
        }
        atoms_chain = set(chain.atom_idxs)

        mmcif_content = (
            f"data_{name}_{cid}"
            + mmcif_write_block(
                "entity",
                ["id", "type", "pdbx_description"],
                [(entity.id, entity.type, entity.pdbx_description)],
            )
            + mmcif_write_block(
                "struct_asym",
                ["id", "entity_id"],
                [(cid, chain.entity_id)],
            )
            + mmcif_write_block(
                "atom_site",
                [
                    "id",
                    "type_symbol",
                    "label_atom_id",
                    "label_comp_id",
                    "label_asym_id",
                    "label_seq_id",
                    "auth_seq_id",
                    "pdbx_PDB_ins_code",
                    "Cartn_x",
                    "Cartn_y",
                    "Cartn_z",
                ],
                [
                    (
                        atom.atom_idx,
                        atom.type_symbol,
                        atom.atom_id,
                        atom.comp_id,
                        atom.chain_id,
                        label_seqs[atom.residue_id],
                        atom.residue_id.seq_id,
                        atom.residue_id.ins_code or ".",
                        f"{crd[0]:.3f}",
                        f"{crd[1]:.3f}",
                        f"{crd[2]:.3f}",
                    )
                    for atom, crd in zip(
                        self.atoms_of_chain(chain),
                        self.coords[chain.atom_idxs],
                    )
                ],
            )
            + mmcif_write_block(
                "chem_comp",
                ["id", "type", "mon_nstd_flag"],
                [
                    (
                        comp.id,
                        comp.type,
                        "y" if comp.mon_nstd_flag else ".",
                    )
                    for comp in self.chem_comps.values()
                    if comp.id in ccid_chain
                ],
            )
            + mmcif_write_block(
                "chem_comp_atom",
                [
                    "comp_id",
                    "atom_id",
                    "type_symbol",
                    "pdbx_aromatic_flag",
                    "pdbx_stereo_config",
                ],
                [
                    (
                        comp_id,
                        atom.atom_id,
                        atom.type_symbol,
                        mmcif_bool(atom.pdbx_aromatic_flag),
                        atom.pdbx_stereo_config or "N",
                    )
                    for comp_id, comp in self.chem_comps.items()
                    if comp.id in ccid_chain
                    for atom in comp.atoms
                ],
            )
            + mmcif_write_block(
                "chem_comp_bond",
                [
                    "comp_id",
                    "atom_id_1",
                    "atom_id_2",
                    "value_order",
                    "pdbx_aromatic_flag",
                    "pdbx_stereo_config",
                ],
                [
                    (
                        comp_id,
                        bond.atom_id_1,
                        bond.atom_id_2,
                        mmcif_bond_order(bond.value_order),
                        mmcif_bool(bond.pdbx_aromatic_flag),
                        bond.pdbx_stereo_config or "N",
                    )
                    for comp_id, comp in self.chem_comps.items()
                    if comp.id in ccid_chain
                    for bond in comp.bonds
                ],
            )
            + mmcif_write_block(
                "struct_conn",
                [
                    "conn_type_id",
                    "pdbx_leaving_atom_flag",
                    "ptnr1_label_asym_id",
                    "ptnr1_label_comp_id",
                    "ptnr1_label_atom_id",
                    "ptnr1_label_seq_id",
                    "ptnr1_auth_seq_id",
                    "pdbx_ptnr1_PDB_ins_code",
                    "ptnr2_label_asym_id",
                    "ptnr2_label_comp_id",
                    "ptnr2_label_atom_id",
                    "ptnr2_label_seq_id",
                    "ptnr2_auth_seq_id",
                    "pdbx_ptnr2_PDB_ins_code",
                ],
                [
                    (
                        conn.conn_type,
                        {0: "none", 1: "one", 2: "both"}[
                            conn.leaving_atom_count
                        ],
                        (ptnr1 := self.atoms[conn.src_idx]).chain_id,
                        ptnr1.comp_id,
                        ptnr1.atom_id,
                        label_seqs[ptnr1.residue_id],
                        ptnr1.residue_id.seq_id,
                        ptnr1.residue_id.ins_code or ".",
                        (ptnr2 := self.atoms[conn.dst_idx]).chain_id,
                        ptnr2.comp_id,
                        ptnr2.atom_id,
                        label_seqs[ptnr2.residue_id],
                        ptnr2.residue_id.seq_id,
                        ptnr2.residue_id.ins_code or ".",
                    )
                    for conn in self.connections
                    if conn.src_idx in atoms_chain
                    and conn.dst_idx in atoms_chain
                ],
            )
            + mmcif_write_block(
                "pdbx_entity_branch_link",
                [
                    "entity_id",
                    "entity_branch_list_num_1",
                    "comp_id_1",
                    "atom_id_1",
                    "leaving_atom_id_1",
                    "entity_branch_list_num_2",
                    "comp_id_2",
                    "atom_id_2",
                    "leaving_atom_id_2",
                    "value_order",
                ],
                [
                    (
                        chain.entity_id,
                        branch.ptnr1.seq_id,
                        branch.ptnr1.comp_id,
                        branch.ptnr1.atom_id,
                        branch.ptnr1.leaving_atom_id,
                        branch.ptnr2.seq_id,
                        branch.ptnr2.comp_id,
                        branch.ptnr2.atom_id,
                        branch.ptnr2.leaving_atom_id,
                        mmcif_bond_order(branch.order),
                    )
                    for branch in merged_branches
                ],
            )
        )

        if write_schemes:
            if chain.type.is_polymer:
                mmcif_content += mmcif_write_block(
                    "pdbx_poly_seq_scheme",
                    [
                        "asym_id",
                        "entity_id",
                        "seq_id",
                        "mon_id",
                        "pdb_seq_num",
                        "pdb_strand_id",
                        "pdb_ins_code",
                    ],
                    [
                        (
                            cid,
                            chain.entity_id,
                            seqres.seq_id,
                            seqres.comp_id,
                            seqres.res_id.seq_id if seqres.res_id else "?",
                            cid,
                            (seqres.res_id.ins_code or ".")
                            if seqres.res_id
                            else "?",
                        )
                        for seqres in chain.seqres
                    ],
                )

            mmcif_content += mmcif_write_block(
                "pdbx_branch_scheme",
                [
                    "asym_id",
                    "entity_id",
                    "num",
                    "mon_id",
                    "pdb_seq_num",
                    "pdb_ins_code",
                ],
                [
                    (
                        scheme.asym_id,
                        scheme.entity_id,
                        scheme.seq_id,
                        scheme.mon_id,
                        scheme.pdb_seq_num,
                        scheme.pdb_ins_code or ".",
                    )
                    for scheme in sorted(branch_scheme)
                    if scheme.pdb_seq_num is not None
                ],
            ) + mmcif_write_block(
                "pdbx_nonpoly_scheme",
                [
                    "asym_id",
                    "entity_id",
                    "mon_id",
                    "ndb_seq_num",
                    "pdb_seq_num",
                    "pdb_ins_code",
                ],
                [
                    (
                        scheme.asym_id,
                        scheme.entity_id,
                        scheme.mon_id,
                        scheme.seq_id,
                        scheme.pdb_seq_num,
                        scheme.pdb_ins_code or ".",
                    )
                    for scheme in sorted(nonpoly_scheme)
                    if scheme.pdb_seq_num is not None
                ],
            )

        return mmcif_content + "\n#\n"

    def to_pdb(self) -> str:
        if not np.all(np.isfinite(self.coords)) or np.any(
            (-1e3 >= self.coords) | (self.coords >= 1e4)
        ):
            raise ValueError("Coordinate value out of range")

        chain_map: dict[str, str] = {}
        chain_ids = list(self.chains.keys())
        chain_ids_avail = dict.fromkeys(
            ascii_uppercase + ascii_lowercase + digits,
            True,
        )
        if len(chain_ids_avail) < len(chain_ids):
            raise ValueError(
                (
                    "Cannot assign unique chain IDs for PDB output "
                    f"for {len(chain_ids)} chains"
                )
            )

        cid_need_remap: list[str] = []
        for cid in chain_ids:
            pdb_cid = cid[:1]
            if chain_ids_avail.pop(pdb_cid, False):
                chain_map[cid] = pdb_cid
            else:
                cid_need_remap.append(cid)

        for cid, pdb_cid in zip(cid_need_remap, chain_ids_avail):
            chain_map[cid] = pdb_cid

        def _resolve_atom_names(atom_id: str, element: str) -> str:
            if (
                len(element) == 1
                and len(atom_id) < 4
                and atom_id.upper().startswith(element.upper())
            ):
                return f" {atom_id:<3}"
            return atom_id.ljust(4)[:4]

        residue_atom_names: dict[ResidueId, dict[str, int]] = defaultdict(dict)

        atoms = sorted(self.atoms, key=lambda a: (a.residue_id, a.atom_idx))
        for atom in atoms:
            res_atom_names = residue_atom_names[atom.residue_id]
            atom_name = _resolve_atom_names(atom.atom_id, atom.type_symbol)
            idx = res_atom_names.setdefault(atom_name, atom.atom_idx)
            if idx != atom.atom_idx:
                raise ValueError(
                    (
                        f"Duplicate atom name {atom_name!r} "
                        f"in residue {atom.residue_id}"
                    )
                )

        atom_names = {
            i: name
            for atom_names in residue_atom_names.values()
            for name, i in atom_names.items()
        }

        def _record_name_for(
            residue_chem_comp: ChemComp,
            atom_chain_type: MolType,
        ) -> str:
            if atom_chain_type in (MolType.RNA, MolType.DNA):
                return "ATOM  "

            if residue_mol_type(residue_chem_comp) != MolType.Protein:
                return "HETATM"

            comp_id = residue_chem_comp.id
            is_standard = comp_id in aa_restype_3to1
            if not is_standard:
                return "HETATM"

            if residue_chem_comp.mon_nstd_flag:
                return "HETATM"

            return "ATOM  "

        chain_atoms: dict[str, list[_PdbAtom]] = defaultdict(list)
        for atom in atoms:
            record = _record_name_for(
                self.residues[atom.residue_id].chem_comp,
                self.chains[atom.chain_id].type,
            )

            ch = chain_map[atom.chain_id]
            rid = ResidueId(
                ch,
                atom.residue_id.seq_id,
                atom.residue_id.ins_code,
            )
            res_name = atom.comp_id[:3]
            atom_name = atom_names[atom.atom_idx]

            chain_atoms[ch].append(
                _PdbAtom(
                    record=record,
                    res_id=rid,
                    name=atom_name,
                    res_name=res_name,
                    coords=self.coords[atom.atom_idx],
                    element=atom.type_symbol,
                    occupancy=atom.occupancy,
                    b_factor=atom.b_factor,
                )
            )

        lines: list[str] = []
        lines.append("REMARK 950 CHAIN ID MAP (label_asym_id -> PDB chain)")
        for cid, one in chain_map.items():
            lines.append(f"REMARK 950   {one} = {cid[:66]}")

        serial = 1
        for ch_atoms in chain_atoms.values():
            for ca in ch_atoms:
                lines.append(ca.to_pdb_line(serial))
                serial += 1

            last_atom = ch_atoms[-1]
            res_name = last_atom.res_name
            res_id = last_atom.res_id
            lines.append(
                (
                    #                  1
                    #                  234567
                    f"TER   {serial:5d}      {res_name:>3} "
                    f"{res_id.chain_id}{res_id.seq_id:4d}{res_id.ins_code:1}"
                )
            )
            serial += 1

            if serial > 99999 or serial < -9999:
                raise ValueError("PDB serial number out of range")

        lines.append("END\n")

        return "\n".join(lines)

    def to_fasta(self, name: str) -> str:
        fastas: list[str] = []
        for chain_id, chain in self.chains.items():
            if not chain.type.is_polymer:
                continue

            seq_list = []
            for seqres in chain.seqres:
                comp_id = seqres.comp_id

                if chain.type == MolType.Protein:
                    comp_id = modres.get(comp_id, comp_id)
                    seq_list.append(aa_restype_3to1.get(comp_id, "X"))
                elif chain.type == MolType.DNA:
                    seq_list.append(dna_restype_3to1.get(comp_id, "N"))
                elif chain.type == MolType.RNA:
                    seq_list.append(rna_restype_3to1.get(comp_id, "N"))

            seq = "".join(seq_list)
            fastas.append(f">{name}_{chain_id}\n{seq}\n")
        return "".join(fastas)


def _resolve_residue_altloc_consistent(altlocs: dict[str, list[AtomSite]]):
    best_atom = max(
        (atom for atoms in altlocs.values() for atom in atoms),
        key=lambda a: a.occupancy,
    )
    comp_id = best_atom.label_comp_id

    for atoms in altlocs.values():
        consistent = [atom for atom in atoms if atom.label_comp_id == comp_id]
        if not consistent:
            continue

        yield max(consistent, key=lambda a: a.occupancy)


def _select_altlocs(atom_sites: list[AtomSite]):
    altlocs: dict[tuple[str, int, str | None], dict[str, list[AtomSite]]] = (
        defaultdict(lambda: defaultdict(list))
    )
    for atom in atom_sites:
        altlocs[atom.label_asym_id, atom.auth_seq_id, atom.pdbx_PDB_ins_code][
            atom.label_atom_id
        ].append(atom)

    return [
        atom
        for residue_altlocs in altlocs.values()
        for atom in _resolve_residue_altloc_consistent(residue_altlocs)
    ]


def residue_mol_type(chem_comp: ChemComp) -> MolType:
    chem_type = chem_comp.type.lower()

    if "peptide" in chem_type:
        return MolType.Protein

    if "rna" in chem_type:
        return MolType.RNA
    if "dna" in chem_type:
        return MolType.DNA

    return MolType.Ligand


def polymer_mol_type(
    seq_infos: list[Scheme],
    chem_comps: dict[str, ChemComp],
) -> MolType:
    molecule_types = {
        residue_mol_type(cc)
        for seq_info in seq_infos
        if (cc := chem_comps.get(seq_info.mon_id)) is not None
    }

    if MolType.Protein in molecule_types:
        return MolType.Protein

    if MolType.RNA in molecule_types:
        return MolType.RNA
    if MolType.DNA in molecule_types:
        return MolType.DNA

    raise ValueError(f"Invalid polymer chain type: {molecule_types}")


def mmcif_chain_types(
    metadata: Mmcif, ccd: dict[str, ChemComp]
) -> dict[str, MolType]:
    chain_types: dict[str, MolType] = {}

    for chain_id, seq_infos in metadata.pdbx_poly_seq_scheme.items():
        chain_types[chain_id] = polymer_mol_type(seq_infos, ccd)

    for chain_id in metadata.pdbx_branch_scheme:
        chain_types[chain_id] = MolType.Ligand

    for chain_id in metadata.pdbx_nonpoly_scheme:
        chain_types[chain_id] = MolType.Ligand

    return chain_types


def _map_struct_conn(
    conn: StructConn,
    chains: dict[str, Chain],
    residues: dict[ResidueId, Residue],
    atoms: list[AssemblyAtom],
) -> AssemblyConnection | None:
    idxs: list[int] = []

    for ptnr in (conn.ptnr1, conn.ptnr2):
        if ptnr.label_asym_id not in chains:
            return None

        residue = residues.get(
            ResidueId(
                ptnr.label_asym_id,
                ptnr.auth_seq_id,
                ptnr.pdbx_PDB_ins_code or "",
            ),
            None,
        )
        if residue is None or residue.chem_comp.id != ptnr.label_comp_id:
            return None

        for atom_idx in residue.atom_idxs:
            if atoms[atom_idx].atom_id == ptnr.label_atom_id:
                idxs.append(atom_idx)
                break
        else:
            return None

    return AssemblyConnection(
        src_idx=idxs[0],
        dst_idx=idxs[1],
        conn_type=conn.conn_type_id.lower(),
        leaving_atom_count=conn.pdbx_leaving_atom_flag,
    )


def _map_branches(
    br_schemes: list[Scheme],
    br_links: list[BranchLink],
    residues: dict[ResidueId, Residue],
):
    schemes = {(scheme.seq_id, scheme.mon_id): scheme for scheme in br_schemes}

    branches = []
    for link in br_links:
        ptnr1 = BranchPartner.from_branch_scheme(
            schemes[(link.ptnr1.entity_branch_list_num, link.ptnr1.comp_id)],
            link.ptnr1,
        )
        ptnr2 = BranchPartner.from_branch_scheme(
            schemes[(link.ptnr2.entity_branch_list_num, link.ptnr2.comp_id)],
            link.ptnr2,
        )
        if (
            residues[ptnr1.res_id].chem_comp.id != ptnr1.comp_id
            or residues[ptnr2.res_id].chem_comp.id != ptnr2.comp_id
        ):
            continue

        branches.append(Branch(ptnr1, ptnr2, link.value_order))

    return branches


def _select_het_seq_scheme(
    schemes: list[Scheme],
    residues: dict[ResidueId, Residue],
):
    seqres: list[SequenceToResidue] = []

    for scheme in schemes:
        s2r = SequenceToResidue.from_scheme(scheme)
        if (
            s2r.res_id is not None
            and (res := residues.get(s2r.res_id, None)) is not None
            and res.chem_comp.id != s2r.comp_id
        ):
            continue

        seqres.append(s2r)

    return seqres


def _model_assembly(
    metadata: Mmcif, atom_sites: list[AtomSite], ccd: dict[str, ChemComp]
) -> Assembly:
    atom_sites = _select_altlocs(atom_sites)

    atom_sites = [
        atom_site for atom_site in atom_sites if atom_site.label_comp_id in ccd
    ]

    coords = np.stack([atom_site.cartn for atom_site in atom_sites])

    atoms = [
        AssemblyAtom(
            atom_idx=i,
            residue_id=ResidueId(
                chain_id=atom_site.label_asym_id,
                seq_id=atom_site.auth_seq_id,
                ins_code=atom_site.pdbx_PDB_ins_code or "",
            ),
            type_symbol=atom_site.type_symbol,
            atom_id=atom_site.label_atom_id,
            comp_id=atom_site.label_comp_id,
            occupancy=atom_site.occupancy,
            b_factor=atom_site.b_iso_or_equiv,
        )
        for i, atom_site in enumerate(atom_sites)
    ]

    residues: dict[ResidueId, Residue] = {}
    chains: dict[str, Chain] = {}

    chain_types = mmcif_chain_types(metadata, ccd)

    for atom in atoms:
        chain = chains.setdefault(
            atom.chain_id,
            Chain(
                atom.chain_id,
                metadata.struct_asym[atom.chain_id],
                chain_types[atom.chain_id],
            ),
        )

        if (residue := residues.get(atom.residue_id, None)) is None:
            residue = residues[atom.residue_id] = Residue(
                atom.residue_id, ccd[atom.comp_id]
            )

            chain.residue_ids.append(residue.residue_id)

        chain.atom_idxs.append(atom.atom_idx)
        residue.atom_idxs.append(atom.atom_idx)

    for cid, chain in chains.items():
        chain.seqres = _select_het_seq_scheme(
            metadata.pdbx_poly_seq_scheme.get(cid, [])
            + metadata.pdbx_branch_scheme.get(cid, [])
            + metadata.pdbx_nonpoly_scheme.get(cid, []),
            residues,
        )
        chain.branches = _map_branches(
            metadata.pdbx_branch_scheme.get(cid, []),
            metadata.pdbx_entity_branch_link.get(chain.entity_id, []),
            residues,
        )
        chain.sync_mappings()

    connections = [
        ac
        for conn in metadata.struct_conn
        if (ac := _map_struct_conn(conn, chains, residues, atoms)) is not None
    ]

    return Assembly(
        metadata=metadata.metadata(),
        coords=coords,
        atoms=atoms,
        residues=residues,
        chains=chains,
        entities=metadata.entity.copy(),
        connections=connections,
    )


def _prepare_initial_assemblies(
    metadata: Mmcif, ccd: dict[str, ChemComp]
) -> list[Assembly]:
    atom_sites_model: dict[int, list[AtomSite]] = defaultdict(list)
    for atom_site in metadata.atom_site:
        if atom_site.type_symbol in "HD":
            continue

        atom_sites_model[atom_site.pdbx_PDB_model_num].append(atom_site)

    return [
        _model_assembly(metadata, atom_sites, ccd)
        for atom_sites in atom_sites_model.values()
    ]


def _smallest_assembly_gens(bas: list[BioAssembly]):
    min_oli_cnt = min(bas, key=lambda ba: ba.oligomeric_count).oligomeric_count
    return [ba for ba in bas if ba.oligomeric_count == min_oli_cnt]


def _first_assembly_gens(bas: list[BioAssembly], assembly_id: int):
    return [ba for ba in bas if ba.id == assembly_id]


def _select_assembly_gens(
    bas: list[BioAssembly], assembly_id: int | None = None
):
    return (
        _first_assembly_gens(bas, assembly_id)
        if assembly_id is not None
        else _smallest_assembly_gens(bas)
    )


def expand_ops(operations: list[list[str]], sym_ops: dict[str, SymOp]):
    return [
        sum(
            (Transformation.from_symop(sym_ops[op]) for op in combi),
            Transformation.ident,
        )
        for combi in itertools.product(*operations)
    ]


def _apply_transforms(
    assem: Assembly,
    generators: list[tuple[list[str], list[Transformation]]],
) -> Assembly:
    assemblies = [
        assem.filter_chains(chain_ids).apply(xforms)
        for chain_ids, xforms in generators
    ]
    return Assembly.join(assemblies)


def mmcif_assemblies(
    data: Mmcif, ccd: dict[str, ChemComp], assembly_id: int | None = None
) -> list[Assembly]:
    assemblies = _prepare_initial_assemblies(data, ccd)
    raw_assemblies = _select_assembly_gens(
        data.pdbx_struct_assembly, assembly_id
    )

    operation_groups = [
        [
            (
                g.asym_id_list,
                expand_ops(g.operations, data.pdbx_struct_oper_list),
            )
            for g in ra.assembly_gens
        ]
        for ra in raw_assemblies
    ]

    return [
        _apply_transforms(initial_assembly, ops)
        for initial_assembly in assemblies
        for ops in operation_groups
    ]



from pathlib import Path

import pytest
from pydantic import TypeAdapter

from gmol.base.data.mmcif import (
    Assembly,
    ChemComp,
    MolType,
    load_components,
    load_mmcif_single,
    mmcif_assemblies,
)


@pytest.fixture(scope="module")
def ccd_with_ligands(test_data: Path) -> dict[str, ChemComp]:
    ccd = load_components(test_data / "ccd" / "components_stdres.cif")
    extra = TypeAdapter(dict[str, ChemComp]).validate_json(
        (test_data / "ccd" / "components_5m7m.json").read_bytes()
    )
    ccd.update(extra)
    return ccd


@pytest.fixture(scope="module")
def assembly_with_ligands(
    test_data: Path, ccd_with_ligands: dict[str, ChemComp]
) -> Assembly:
    data = load_mmcif_single(test_data / "mmcif" / "5m7m.cif")
    assemblies = mmcif_assemblies(data, ccd_with_ligands)
    assert len(assemblies) == 1
    return assemblies[0]


def test_ligand_label_seq_id_is_dot(assembly_with_ligands: Assembly):
    """Non-polymer (ligand) chains should have '.' for label_seq_id in mmCIF output."""
    mmcif_str = assembly_with_ligands.to_mmcif("test")

    # Parse atom_site lines from the mmCIF output
    in_atom_site = False
    field_names: list[str] = []
    label_seq_id_col = -1
    label_asym_id_col = -1

    ligand_chain_ids = {
        cid
        for cid, chain in assembly_with_ligands.chains.items()
        if chain.type == MolType.Ligand
    }
    polymer_chain_ids = {
        cid
        for cid, chain in assembly_with_ligands.chains.items()
        if chain.type.is_polymer
    }

    assert len(ligand_chain_ids) > 0, "Test requires ligand chains"
    assert len(polymer_chain_ids) > 0, "Test requires polymer chains"

    ligand_label_seq_ids: set[str] = set()
    polymer_label_seq_ids: set[str] = set()

    for line in mmcif_str.splitlines():
        line = line.strip()

        if line == "loop_":
            in_atom_site = False
            field_names = []
            continue

        if line.startswith("_atom_site."):
            field_names.append(line.split(".")[1])
            if field_names[-1] == "label_seq_id":
                label_seq_id_col = len(field_names) - 1
            if field_names[-1] == "label_asym_id":
                label_asym_id_col = len(field_names) - 1
            in_atom_site = True
            continue

        if (
            in_atom_site
            and label_seq_id_col >= 0
            and line
            and not line.startswith("_")
            and not line.startswith("#")
        ):
            tokens = line.split()
            if len(tokens) >= len(field_names):
                asym_id = tokens[label_asym_id_col]
                seq_id = tokens[label_seq_id_col]

                if asym_id in ligand_chain_ids:
                    ligand_label_seq_ids.add(seq_id)
                elif asym_id in polymer_chain_ids:
                    polymer_label_seq_ids.add(seq_id)

        if line.startswith("#") and in_atom_site:
            in_atom_site = False

    # Ligand label_seq_id must all be "."
    assert ligand_label_seq_ids == {"."}, (
        f"Ligand label_seq_id should be '.', got {ligand_label_seq_ids}"
    )

    # Polymer label_seq_id must be numeric (not ".")
    assert "." not in polymer_label_seq_ids, (
        "Polymer label_seq_id should not contain '.'"
    )
    assert len(polymer_label_seq_ids) > 0
